import os
import sys
import json
import logging
from datetime import datetime
import platform
import distro
import getpass
import logging
from collections import OrderedDict
import logging

from enrichsdk.utils import SafeEncoder

from . import log

logger = logging.getLogger("app")


#######################################################
# Config object to handle a number of helper functions
#######################################################
class Config(object):
    """
    Context object having a number of capabilities including
    log management, files and path management, credentials
    lookup etc.

    """

    def __init__(self, params, *args, **kwargs):
        """
        Initialize the config (context) object.

        Args:
          params (dict): Only two keys are supported. debug (bool) prints
                         extra information. It is useful for debugging loading.
                         settings is the path to the settings file.
        """
        super().__init__(*args, **kwargs)
        self.debug = params.get("debug", False)
        self.settings = params.get("settings", None)
        self.readonly = params.get("readonly", False)
        self.show_metadata = params.get("show_metadata", True)

        if self.settings is not None:
            self.settings = os.path.expanduser(self.settings)

        self.now = datetime.now()
        self.conf = {"credentials": {}, "libraries": []}
        self.stats = {
            "platform": {
                "node": platform.node(),
                "os": platform.system(),
                "release": platform.release(),
                "processor": platform.processor(),
                "python": platform.python_version(),
                "distribution": distro.linux_distribution(),
            },
            "user": getpass.getuser(),
        }

        self.state = self.get_default_metadata()
        self.state.update({"outputs": [], "notes": [], "message": ""})

    def get_default_metadata(self):
        """
        Get reusable metadata dict with some standard fields.
        """
        metadata = {
            "schema": "standalone:command",
            "version": "1.0",
            "timestamp": datetime.now().replace(microsecond=0).isoformat(),
            "command": {
                "name": "Unknown",
                "status": "initial",
                "description": "",
                "host": self.stats["platform"]["node"],
                "stats": self.stats,
                "pid": os.getpid(),
                "start_time": self.now,
                "end_time": self.now,
                "cmdline": list(sys.argv),
            },
        }

        return metadata

    def configure(self, settings=None):
        """
        Configure the config object

        Args:
           settings (str): Path to the settings file

        """
        if settings is None:
            settings = self.settings

        if (settings is None) or (not os.path.exists(settings)):
            # Quietly return. Do nothing..
            raise Exception("Settings path invalid or undefined: {}".format(settings))

        try:
            loaded_conf = json.load(open(settings))
        except:
            logger.exception("Invalid settings file")
            loaded_conf = None

        if loaded_conf is None:
            return

        # Materialize the paths...
        self.enrich_root = loaded_conf.get("enrich_root")
        self.enrich_root = os.path.expanduser(self.enrich_root)
        self.enrich_root = os.path.abspath(self.enrich_root)

        libraries = loaded_conf.get("libraries", [])
        libraries = [self.get_file(l) for l in libraries]
        self.conf["libraries"] = libraries

        # Insert the rest
        defaults = {
            "enrich_data": "%(enrich_root)s/data",
            "enrich_customers": "%(enrich_root)s/customers",
            "runid": "run-%Y%m%d-%H%M%S-%(user)s",
            "command_root": "%(enrich_data)s/_commands",
            "rundir": "%(command_root)s/%(hostname)s/%(runid)s",
            "outputdir": "%(rundir)s/output",
            "log": "%(rundir)s/log.json",
            "metadata": "%(rundir)s/metadata.json",
            "args": {},
        }

        for name in [
            "enrich_data",
            "enrich_customers",
            "runid",
            "command_root",
            "rundir",
            "outputdir",
            "log",
            "metadata",
        ]:
            default = defaults[name]
            try:
                final = loaded_conf.get(name, default)
                final = self.get_file(final)
            except:
                error = "Could not resolve: {} ({})".format(name, final)
                raise Exception(error)

            setattr(self, name, final)
            self.state["command"][name] = final

        for name in ["args"]:
            default = defaults[name]
            try:
                final = loaded_conf.get(name, default)
            except:
                error = "Could not resolve: {} ({})".format(name, final)
                raise Exception(error)

            setattr(self, name, final)

        # Set environment variables.....
        os.environ["ENRICH_ROOT"] = self.enrich_root
        os.environ["ENRICH_CUSTOMERS"] = self.enrich_customers
        os.environ["ENRICH_DATA"] = self.enrich_data

    def get_file(self, path):
        """
        Resolve the abstract specification into a full path

        Args:
           path (str): Path specification

        Example::

           %(enrich_data)/acme/Projects/commands

        This is resolved into `/home/ubuntu/enrich/data/acme/Projects/commands`

        """
        path = path.replace("%(", "%%(")  # This will make the strftime handling safe
        path = self.now.strftime(path)
        path = path % {
            "enrich_root": self.enrich_root,
            "enrich_data": getattr(self, "enrich_data", ""),
            "enrich_customers": getattr(self, "enrich_customers", ""),
            "user": self.stats["user"],
            "username": self.stats["user"],
            "hostname": self.stats["platform"]["node"],
            "node": self.stats["platform"]["node"],
            "runid": getattr(self, "runid", ""),
            "rundir": getattr(self, "rundir", ""),
            "outputdir": getattr(self, "outputdir", ""),
        }
        return path

    def get_relative_path(self, path, what="enrich_root"):

        if what == "enrich_root":
            return os.path.relpath(path, self.enrich_root)
        elif what == "enrich_data":
            return os.path.relpath(path, self.enrich_data)
        else:
            raise Exception("relative path should specify the root")

    def add_output(self, params):
        """
        Add file to the command output

        Args:
           params (dict): file parameters
              File parameters includes a name of the output, relative path
              to the file, and a description for the output

        """
        required = ["name", "description", "path"]
        missing = [n for n in required if n not in params]
        if len(missing) > 0:
            raise Exception("Missing parameters for the output: {}".format(missing))

        if os.path.isabs(params["path"]):
            raise Exception(
                "Path specified for output {} should be relative".format(params["name"])
            )

        for o in self.state["outputs"]:
            if o["name"] == params["name"]:
                raise Exception(
                    "An output already exists with the name: {}".format(params["name"])
                )

        params["fullpath"] = self.get_file("%(outputdir)s/{}".format(params["path"]))
        params["relpath"] = self.get_relative_path(params["fullpath"])

        # Now append
        self.state["outputs"].append(params)

    def get_output_handle(self, name, *args):
        """
        Get a file descriptor to the named output.

        Args:
           name (str): Name of the output.
              This output was previously added using `add_output` call.

        """
        for o in self.state["outputs"]:
            if o["name"] == name:
                try:
                    os.makedirs(os.path.dirname(o["fullpath"]))
                except:
                    pass
                return open(o["fullpath"], *args)

        raise Exception("Could not find output {}".format(name))

    def get_library_paths(self):
        return self.conf["libraries"]

    def get_args(self, cmd):
        """
        Get default arguments to the command specified in the settings file

        Args:
           cmd (str): Name of the command

        """
        logger = logging.getLogger("app")
        if not hasattr(self, "args"):
            if self.debug:
                logger.error("args not present")
            return {}

        if not isinstance(self.args, dict):
            if self.debug:
                logger.error("args not a dictionary")
            return {}

        return self.args.get(cmd, {})

    # Update the state
    def set_status(self, value):
        self.state["command"]["status"] = value

    def set_message(self, value):
        self.state["message"] = value

    def dump_state(self):

        # Update file statistics
        for o in self.state["outputs"]:
            path = o["fullpath"]
            st = os.stat(path)
            o["size"] = st.st_size

        metadata = OrderedDict(
            [
                ("schema", self.state["schema"]),
                ("version", self.state["version"]),
                ("timestamp", self.state["timestamp"]),
                ("command", self.state["command"]),
                ("message", self.state["message"]),
                ("outputs", self.state["outputs"]),
                ("notes", self.state["notes"]),
                ("log", os.path.relpath(self.log, self.rundir)),
            ]
        )

        if not self.readonly:
            with open(self.metadata, "w") as fd:
                json.dump(metadata, fd, indent=4, cls=SafeEncoder)
        elif self.show_metadata:
            print(json.dumps(metadata, indent=4, cls=SafeEncoder), file=sys.stderr)

        if self.debug:
            logger.debug("Dumped execution metadata")

    # Logging start and end
    def start_run(self, cmd):

        # Override with parameters from settings
        if not self.readonly:
            args = self.get_args(cmd)
            self.readonly = args.get("readonly", self.readonly)
            self.show_metadata = args.get("show_metadata", self.show_metadata)

        # Setup logging
        if self.readonly:
            log.logging_start()
        else:
            log.logging_start(self.log)

        logger = logging.getLogger("app")
        if self.debug:
            logger.debug("Starting run")

        # Dump the state at the beginning
        self.dump_state()

    def end_run(self):

        self.state["command"]["end_time"] = datetime.now()
        self.dump_state()

        # Dump status
        logger = logging.getLogger("app")

        if self.debug:
            logger.debug("Ending run")
        log.logging_end()



import os
import json
import pandas as pd
from langchain.llms import OpenAI
from langchain import PromptTemplate

__all__ = ['SADLClassifier']

class SADLClassifier(object):
    """
    Class to take in a dataframe and create a data dictionary for it
    """

    def __init__(self,
                 cred,
                 df,
                 context="",
                 entities=None):
        """
        init SADL object
        cred: credentials object
        df: dataframe
        context: any useful context when doing the labelling
                    usually, this will include keywords that identify
                    the industry/domain that the dataframe is from
        entities: path to a json spec with any additional entity mappings
                    needed by the labeller
        """

        # init
        self.cred = cred
        self.df = df
        self.context = context
        self.model = "openai"

        # init the base entity map
        self.entity_map = {
            "person": ["id", "name", "gender", "occupation", "nationality", "age", "marital status", "education level", "salutation", "job title", "relationship", "other"],
            "location": ["id", "city", "state", "province", "country", "address", "postal code", "latitude and longitude", "landmark", "neighborhood", "region", "time zone", "other"],
            "organization": ["id", "type", "industry", "other"],
            "event": ["id", "type", "theme", "other"],
            "product": ["id", "category", "brand", "model", "feature", "other"],
            "service": ["id", "category", "cost", "feature", "other"],
            "date/time": ["id", "date", "day", "month", "year", "timestamp", "time zone", "duration", "other"],
            "identifier": ["id"],
            "metric": ["revenue", "size", "volume", "sales", "measurement", "price", "other"],
            "contact": ["website", "email", "phone number", "social link"]
        }

        # init the base industry list
        self.industries = ["Business", "Technology",
                           "Finance", "Healthcare",
                           "Manufacturing", "Retail",
                           "Government", "eCommerce",
                           "Education", "Logistics"]

        # load additional entities if needed
        if entities:
            self.entity_map = self.load_entities(entities)

        # get the LLM obj
        self.llm = self._get_llm_objs(model=self.model,
                                      cred=self.cred)


    ## helper functions

    def _get_llm_objs(self, model, cred):
        """
        setup the LLM
        """
        # get the api key from creds
        api_key = self._get_api_key(cred)

        # init the model
        if model == "openai":
            # get the llm object
            llm = OpenAI(temperature=0,
                         max_tokens=1024,
                         openai_api_key=api_key)
        else:
            llm = None

        return llm

    def _get_api_key(self, cred):
        """
        get the API key from the cred
        """
        api_key = None
        if isinstance(cred, str):
            api_key = cred
        if isinstance(cred, dict) and 'apikey' in cred:
            api_key = cred['apikey']
        return api_key


    ## interfaces

    def load_entities(self, entities):
        """
        load any additional entity mappings and add them to the
        base entity map
        entities: path to entity mappings
        """

        # get the base entity map
        entity_map = self.entity_map

        # check path
        if os.path.exists(entities) and entities.endswith(".json"):
            with open(entities, "r") as fd:
                addl_emap = json.load(fd)
            if isinstance(addl_emap, dict):
                for  e, se in addl_emap.items():
                    # lowercase everything
                    e = e.lower()
                    # we need lists
                    if isinstance(se, str):
                        se = [se.lower()]
                    # lowercase everything
                    if isinstance(se, list):
                        se = [i.lower() for i in se]
                    else:
                        # we cannot add this item, move on
                        continue
                    # we can now add this mapping to the base map
                    entity_map[e].extend(se)
                    # make sure we remove duplicates
                    entity_map[e] = list(set(entity_map[e]))
            # done, we have the updated map
            self.entity_map = entity_map
        else:
            pass

        return self.entity_map


    def generate_prompt_columns(self, df, context=""):
        """
        generate a prompt for labelling a dataframe given some context
        """

        def construct_promptpartial_from_map(entity_map):
            """
            Take an entity map and construct the entity mapping partial
            that goes into the prompt for the LLM
            """
            # first the entities
            entities = ", ".join(list(entity_map.keys()))
            partial_e = f"Assume the entities available for classification are as follows: \n{entities}"

            # then for the sub-entities
            partial_se = ""
            for e, se in entity_map.items():
                p_se = f"Assume the following are the sub entities for {e}: \n{', '.join(se)}"
                partial_se = f"{partial_se}\n\n{p_se}"

            partial = f"{partial_e}{partial_se}"

            return partial

        def construct_promptpartial_outputformat():
            """
            Construct the output format structure
            for the response from the LLM
            """
            column_label = "column_name"
            label_fields = ['datatype', 'entity', 'sub_entity', 'pii', 'description']

            p = ", ".join([f'"{f}": {f}' for f in label_fields])
            op_format = "{" + column_label + ": {" + p + "}}"

            return op_format


        # construct the prompt partial for the entity mappings
        map_partial = construct_promptpartial_from_map(self.entity_map)

        # get the input column name labels
        labels = ", ".join(list(df.columns))

        # output format
        output_format = construct_promptpartial_outputformat()

        # setup the prompt
        template = """You are a data entry operator.
Your task is to construct a data dictionary for a set of database column names given to you.
The database is from a company in {context} industry.

{map_partial}

Classify the following column names into entity and sub_entity using the above mentioned details.
Also provide the following for each column:
- datatype
- true/false indicating whether the column could contain personally identifiable information (PII)
- description

Format your output as a nested json dictionary as follows:
{output_format}

Here are the input column names:
{labels}"""

        prompt = PromptTemplate(
            input_variables=["context", "map_partial", "labels", "output_format"],
            template=template
        )

        return prompt.format(context=context, map_partial=map_partial, labels=labels, output_format=output_format)

    def generate_prompt_industry(self, df):
        """
        generate a prompt for identifying the industry of a dataframe given some context
        """
        # get the input labels
        labels = ", ".join(list(df.columns))

        # setup the prompt
        template = """You are a data entry operator.
Assume you have a list of industries:
{industries}

Your task is to identify what industry a dataset belongs to given the columns in the dataset.
Respond with EXACTLY one option from the list of industries.

Here are the columns in the dataset:
{labels}"""

        prompt = PromptTemplate(
            input_variables=["labels", "industries"],
            template=template
        )

        return prompt.format(labels=labels, industries=self.industries)


    def classify_columns(self, context=""):
        """
        label the columns in the dataframe
        """

        # get the dataframe
        df = self.df

        # set the context for this labelling attempt
        context = self.context if context=="" else context

        # get the prompt for the LLM
        prompt = self.generate_prompt_columns(df, context)

        # prompt the LLM
        response = self.llm(prompt)

        try:
            result = json.loads(response.lower().strip())
            result_type = 'json'
        except:
            result = response.strip()
            result_type = 'str'

        # process the response
        return result, result_type

    def classify_industry(self):
        """
        figure out what industry the dataframe is from
        """

        # get the dataframe
        df = self.df

        # get the prompt for the LLM
        prompt = self.generate_prompt_industry(df)

        # prompt the LLM and get the response
        response = self.llm(prompt)
        result = response.lower().strip()

        # process the response
        return result



if __name__ == "__main__":

    cred = get_credentials_by_name('openai-api')

    # Sample file
    csv_path = "revenue.dashboard_market.csv"
    emap_file = "emap.json"

    # Access other data from here and specify local path
    #s3path      = "s3://scribble-demodata/sadl/demo-server/sadl-demo/sadl-llm"
    #csv_file    = f"{s3path}/revenue.dashboard_market.csv"
    #emap_file   = f"{s3path}/emap.json"

    # load csv and make df
    df = pd.read_csv(csv_file)

    # add two new dummy columns
    # only to show off SADL capabilities
    df['txn_id']            = df.index
    df['sender_fullname']   = "dummy"

    # get a SADL agent
    labeller = SADLClassifier(cred=cred,
                                df=df,
                                entities=emap_file)

    # -- optional step --
    # get a label for possible industry
    # not needed if context param is specified when creating a new SADL object
    industry = labeller.classify_industry()

    # classify the column labels
    result, result_type = labeller.classify_columns(context=industry)
    if result_type=='json':
        print (json.dumps(result, indent=4))
    else:
        print (result)



import io
import os
import json
import math
import pickle
import numpy as np
import pandas as pd
import time
import sqlite3
import tempfile
import hashlib
from enrichsdk import Compute, S3Mixin
from datetime import datetime, timedelta, date
import logging
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.patches import Rectangle

from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec
from enrichsdk.contrib.lib.assets import BaseProphetForecasterModel
from enrichsdk.datasets import TransformDoodle
from enrichsdk.utils import SafeEncoder, get_yesterday


from dateutil import parser as dateparser

logger = logging.getLogger("app")

class TimeSeriesForecasterBase(Compute):
    """
    Take a timeseries and project it's future values with exogenous variables.
    Features of transform baseclass include:
        * Flexible configuration
        * Highlevel specification of time series forecasting
            * specified data source or custom method to generate one
            * by default, forecast using facebook's prophet library
             or custom defined ones using other libraries
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "TimeSeriesForecasterBase"
        self.description = "Forecast future values of a timeseries"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

        self.default_strategy = "prophet"
        self.default_type = "vanilla"

        self.epoch = time.time()    #for output path


    @classmethod
    def instantiable(cls):
        return False

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}


    ###########################################
    # Process the forecaster spec
    ###########################################
    def process_spec(self, spec, data):
        """
        Process the forecaster spec.
        generate result and chart for each forecaster
        """

        def store_chart(tsf):
            '''
            We generate multiple charts for each forecaster
            Immeidately store the charts as we generate
            '''
            viz = tsf['viz']
            filename = f"{forecaster_name}-{name}-forecasting.png"
            msg = self.store_viz(spec, filename, viz)
            tsf.pop('viz', None)
            return msg

        msg = ""
        name = spec["name"]
        config = spec["config"]

        forecasters = config['forecasters']

       # if forecasters is not a dict
        if not isinstance(forecasters, dict):
            logger.exception("Forecasters must be a dict",
                             extra={"transform": self.name})
            raise Exception("Forecasters must be a dict")

        forecasters = config['forecasters']

        result = {"forecasts": {}}
        for forecaster_name, forecaster in forecasters.items():

            tsf = self.run_forecasting(spec, data, forecaster_name, forecaster)
            msg+= store_chart(tsf)

            logger.debug(f"Processed and then aved visualization for {forecaster_name}",
                    extra={"transform": self.name, "data": msg})

            result['forecasts'][forecaster_name] = tsf

        logger.debug(f"Done processing all the forecasters",
                extra={"transform": self.name})

        return result


    def run_forecasting(self, spec, data, forecaster_name, forecaster):
        """
        Instantiate the forecaster and run forecasting
        """
        # default is prophet
        # type is vanilla
        strategy = forecaster.get('strategy', self.default_strategy)
        type = forecaster.get('type', self.default_type)
        params = forecaster.get('params', {})

        # return timeseries forecast
        tsf = {}
        chart_params = params.get('chart_params', {})

        if strategy == 'prophet':

            if type == "vanilla":
                observation  = params.get('observation', None)

                if observation is None:
                    logger.exception(f"Observation time series must be specified for forecaster: {forecaster_name}",
                                    extra={"transform": self.name, "data": json.dumps(forecaster, indent=4)})
                    raise Exception("Observation must be specified for prophet forecaster")

                df = data['observations'][observation]

                forecast_obj = BaseProphetForecasterModel(df)
                forecast = forecast_obj.run_forecasting(params)

                viz = forecast_obj.visualize_forecasting(forecast, chart_params)
                del forecast_obj

            elif type ==  "exogenous":
                df = data['combined']

                forecast_obj = BaseProphetForecasterModel(df)
                forecast = forecast_obj.run_forecasting(params)
                viz = forecast_obj.visualize_forecasting(forecast, chart_params)
                del forecast_obj

            else:
                logger.excption(f"Invalid type for prophet forecaster: {forecaster_name}",
                                extra={"transform": self.name, "data": json.dumps(forecaster, indent=4)})
                raise Exception("Invalid type for prophet forecaster")

        tsf = {
            "forecast" : forecast,
            "viz" : viz,
        }
        msg = note(forecast, f"Forecast for {forecaster_name}")
        logger.debug(f"Forecasted time series for {forecaster_name}",
                        extra={"transform": self.name, "data": msg})

        return tsf


    ###########################################
    # Helper Functions
    ###########################################

    def update_frame(self, source, description, df, lineage=None):
        if isinstance(source, str):
            name = source
        else:
            name = source["name"]

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    ###########################################
    # I/O Functions
    ###########################################

    def read_s3_data(self, filename, params={}):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            try:
                df = pd.read_csv(s3.open(filename), **params)
                return df
            except:
                pass
        return None

    def get_dataset_s3(self, spec, source, paths, start_date, end_date):
        '''
        Gets all files from paths and puts them together
        into a single dataframe. If self.args['cache']==True,
        then this consolidated dataframe is cached / read from cache
        as applicable.
        '''
        msg = ""

        run_date    = self.args['run_date']
        config      = spec['config']
        dataset     = source['dataset']
        params      = source.get('params', {})

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{start_date}-to-{end_date}"
        cachefile = f"cache/{self.name}-rawdata-cache-" + cachename + ".csv"

        # read from cache if available
        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile, **params)
                logger.debug(f"Read cached {dataset}", extra={"transform": self.name, "data": msg})
                return df

        # read from S3
        dfs = []
        for path in paths:
            _df = self.read_s3_data(path, params)
            if _df is None:
                msg += f"Path error, skipping: {path}" + "\n"
                continue
            msg += f"Read from path: {path}" + "\n"
            dfs.append(_df)
        df = pd.concat(dfs)

        logger.debug(f"Read fresh {dataset}", extra={"transform": self.name})

        # Cache it for future use
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if (len(paths) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": paths,
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec, f"Dataset: {dataset}", df, lineage)

        return df

    def load_dataset(self, spec, name, source, datewindow):
        msg = ""

        source_id       = source['source_id']
        dataset         = source['dataset']
        source_version  = source.get('source_version', 'v1')

        start_date      = datewindow['start_date']
        end_date        = datewindow['end_date']

        if source_id == "custom":
            # we have a custom defined method in the derived class to generate the dataset
            if not hasattr(self, dataset):
                logger.exception(
                    f"No handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
                return None

            handler = getattr(self, dataset)
            params = source.get("params", {})
            df = handler(start_date, end_date, params)
            if df is None:
                logger.exception(
                    f"Couldn't load dataset for {spec['name']}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
                return None

            msg += f"Loaded using custom method: {dataset}" + "\n"
        else:
            # we are using the SDK to get the dataset
            datacred = self.args['datacred']
            doodle = TransformDoodle(self, self.state, datacred)

            d_source, paths = doodle.get_source_paths(start=start_date,
                                                    end=end_date,
                                                    name=dataset,
                                                    version=source_version,
                                                    source_id=source_id)

            msg += f"Paths: {paths}" + "\n"

            # Insert a read action..
            try:
                result = doodle.update_source(d_source['id'], {})
                logger.debug(f"Updated Doodle: {dataset}",
                             extra={
                                 'transform': self.name,
                                 'data': json.dumps(result, indent=4, cls=SafeEncoder)
                             })
            except:
                logger.exception(f"Unable to update Doodle {dataset}",
                                 extra={
                                     'transform': self.name
                                 })

            # read the data from S3
            df = self.get_dataset_s3(spec, source, paths, start_date, end_date)


        # check if we need to perform a post-processing step on the input data
        postprocess = source.get("postprocess")
        if postprocess != None:
            if not hasattr(self, postprocess['method']):
                logger.exception(
                    f"No post-process handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
            else:
                handler = getattr(self, postprocess['method'])
                params = postprocess.get("params", {})
                df = handler(df, params)
                msg += f"Post-processed dataset={dataset} using handler: {postprocess}" + "\n"

        msg += note(df, f"Input Dataset: {dataset}")
        logger.debug(
            f"Loaded dataset={dataset} for source={name}",
            extra={"transform": self.name, "data": msg}
        )

        return df

    def store_viz(self, spec, filename, viz):
        appname     = spec.get('app', self.name)
        name        = spec['name']
        namespace   = spec.get('namespace', 'default')
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = self.epoch

        msg = ""

        # write to s3
        targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")
        vizfile = os.path.join(targetdir, f"{filename}")

        img_data = io.BytesIO()
        viz.savefig(img_data, format='png')
        img_data.seek(0)
        with s3.open(vizfile, 'wb') as fd:
            fd.write(img_data.getbuffer())

        msg += f"Stored (remote): {vizfile}" + "\n"

        return msg


    def s3_store_result(self, spec, result):
        appname     = spec.get('app',self.name)
        name        = spec['name']
        namespace   = spec.get('namespace', 'default')
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = self.epoch


        # where are we storing it?
        targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")

        metadatafile = os.path.join(targetdir, f"metadata.json")
        metadata = self.get_default_metadata(self.state)
        metadata['spec'] = spec

        ## write to s3
        # forecast dfs
        msg = ""
        for forecaster_name in result[name].get("forecasts", {}):
            resultfile = os.path.join(targetdir, f"forecast_{forecaster_name}.csv")
            msg += f"Forecast for {forecaster_name}: {resultfile}" + "\n"
            with s3.open(resultfile, 'w') as fd:
                result[name]["forecasts"][forecaster_name]['forecast'].to_csv(fd, index=False)

        # extras
        if "extra" in result[name]:
            resultfile = os.path.join(targetdir, f"extra.json")
            msg += f"Extras for {forecaster_name}: {resultfile}" + "\n"
            with s3.open(resultfile, 'w') as fd:
                json.dump(result[name]['extra'], fd, indent=4, cls=SafeEncoder)

        # metadata
        msg += f"Metadata location: {metadatafile}" + "\n"
        with s3.open(metadatafile, 'w') as fd:
            json.dump(metadata, fd, indent=4, cls=SafeEncoder)

        logger.debug(f"Wrote timeseries forecasting result to S3",
                        extra={"transform": self.name,
                                "data": msg})


    def store_result(self, spec, result):
        name    = spec['name']
        config  = spec['config']
        store   = config.get('store', {"sink": "s3"})
        stored  = False

        for f in ["sink"]:
            if f not in store:
                logger.exception(
                    f"Store has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                return

        sink = store['sink']
        if sink == "s3":
            try:
                # store in s3
                self.s3_store_result(spec, result)
                stored = True
            except:
                pass
        else:
            logger.exception(f"Unknown store for dataset: {name}",
                         extra={
                             'transform': self.name
                         })

        # update KPI for this spec
        self.state.update_kpi(f"{name}_ts_forecasted", stored)


    def get_datewindow(self, source, spec):
        """
        Set the time window for observations and exogenous variables.
        Get both of these from args parameters
        if not start_date defaults to 60 days prior to end date
        end_date is day prior to run_date, which is usually today
        """
        datewindow = {}
        default_delta = 60

        run_date = self.args['run_date']

        try:
            if 'end_date' in self.args and self.args['end_date']:
                end_date = datetime.fromisoformat(self.args['end_date'])
            else:
                logger.debug(
                    f"End date not in args. Using yesterday's date.")
                end_date = run_date - timedelta(days=1)

            if 'start_date' in self.args and self.args['start_date']:
                start_date = datetime.fromisoformat(self.args['start_date'])
            else:
                logger.debug(
                    f"Start date not in args. Using {default_delta} days prior to end date. ")
                start_date = end_date - timedelta(days=default_delta)
        except Exception as e:
            logger.exception(
                f"Error parsing date window for {spec['name']}.",
                extra={"transform": self.name, "data": self.args}
            )
            datewindow = None
            return datewindow

        if start_date > end_date:
            logger.exception(
                        f"Start date greater than end date. Skipping the spec {spec['name']}.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
            datewindow = None # be explicit
            return datewindow

        datewindow['start_date'] = start_date
        datewindow['end_date'] = end_date

        return datewindow

    ######################################################################
    ## Helper functions for process
    ######################################################################
    def precheck_spec(self, spec):
        '''
        Check if the spec is valid
        '''
        is_valid_spec = True
        name = spec.get('name', 'NO_SPEC_NAME')

        enabled = spec.get("active", True)
        if not enabled:
            logger.debug(
                f"Spec not enabled, skipping.",
                extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
            )
            is_valid_spec = False
            return is_valid_spec

        for f in ["name", "config"]:
            if f not in spec:
                logger.exception(
                    f"Spec has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                is_valid_spec = False
                return is_valid_spec

        config = spec['config']

        for f in ["source", "forecasters"]:
            if f not in config:
                logger.exception(
                    f"Spec config has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                is_valid_spec = False
                return is_valid_spec

        return is_valid_spec

    def load_source(self, spec):
        """
        Load all the sources to a 'data' dict
        modifies the 'data' dict.
        """
        config = spec['config']
        source = config['source']

        data = {}
        is_valid = True

        if 'observations' not in source:
            logger.exception(
                f"Spec config has no observations param, skipping.",
                extra={"transform": self.name, "data": json.dumps(spec, indent=4)})
            is_valid_spec = False
            return is_valid

        if 'exovars' not in source:
            logger.debug(
                f"Exogenous variables not specified in {spec['name']}",
                extra={"transform": self.name, "data": json.dumps(spec, indent=4)})

        # get time_window for observations and exovars
        datewindow = self.get_datewindow(source, spec)
        if datewindow is None:
            logger.debug(
                f"Invalid date window for {spec['name']}",
                extra={"transform": self.name})
            is_valid_spec = False
            return is_valid, data

        data['observations'] = {}
        for dataname, dataspec in source['observations'].items():
            dataset = self.load_dataset(spec, dataname, dataspec, datewindow)
            data["observations"][dataname] = dataset

        # then load the exovars data set if specified
        if "exovars" in source:
            data['exovars'] = {}
            for dataname, dataspec in source['exovars'].items():
                dataset = self.load_dataset(spec, dataname, dataspec, datewindow)
                data["exovars"][dataname] = dataset

        return is_valid, data

    def combined_dataset(self, spec, data):
        """
        Adds the combined dataset to the data dict
        """
        config = spec['config']
        combined_dataset = pd.DataFrame()

        if "combine_sources" in config:
            combine_sources = config["combine_sources"]
            dataset = combine_sources.get("dataset", None)

            if hasattr(self, dataset):
                params = combine_sources.get("params", {})
                handler = getattr(self, dataset)
                combined_dataset =  handler(params, data, spec)
                data['combined'] = combined_dataset

                msg = note(combined_dataset, "Combined dataset")
                logger.debug(f"Combined dataset for {spec['name']}",
                             extra={"transform": self.name, "data": msg})

        return data

    def postprocess_results(self, spec, result):
        """
        Postprocess the results. The method defined in the subclass
        """
        config = spec['config']
        # do post_process results
        postprocess_results = config.get('postprocess_results', None)
        if postprocess_results:
            method = postprocess_results.get('method', "")
            params = postprocess_results.get('params', {})
            handler = getattr(self, method, None)
            if handler:
                result = handler(spec, result, params)
            else:
                logger.exception(
                    f"Spec: {spec['name']} -- postprocess_results method not found",
                    extra={"transform": self.name}
                )
        logger.debug(f"Postprocess results for {spec['name']} done",
                     extra={"transform": self.name})

        return result


    ###########################################
    # Main function
    ###########################################
    def process(self, state):
        """
        Run the computation and update the state
        1. Load the datasets
        2. Run forecasting
        3. process the forecasting results
        4. store the results
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        # Get the profile spec
        is_valid, profile, msg = profilespec.get_profile(self, "policyapp.forecasting")
        if is_valid:
            logger.debug(
                f"Loaded profilespec",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")
        specs = profile.get("specs", None)
        if specs is None:
            raise Exception("Could not find 'specs' in profile")

        # Now go through each spec and process it
        for spec in specs:

            do_process_spec = self.precheck_spec(spec)
            if do_process_spec == False:
                continue

            ## we can now proceed with processing the spec
            # load source
            do_process_spec, data = self.load_source(spec)
            if do_process_spec == False:
                continue

            # post process the sources
            data = self.combined_dataset(spec, data)

            # run the forecasters
            result = self.process_spec(spec, data)
            if result is None:
                continue

            # postprocess the results
            result = self.postprocess_results(spec, result)

            # tag the result under the spec name
            result = {spec['name']: result}

            # store the  results
            self.store_result(spec, result)


        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



"""

This module implements the DataGPT microservice. Its purpose is to wrap
around the DataGPT agent and provide an integration point from end applications.
The DataGPT module enables users to run queries on datasets.

"""
import os
import sys
import re
import glob
import yaml
import json
import traceback
import time
import hashlib
import struct
import logging
import threading
from uuid import uuid4
from datetime import datetime
from fastapi import FastAPI, BackgroundTasks, Depends
from fastapi.concurrency import run_in_threadpool
from fastapi_utils.tasks import repeat_every

from llmsdk.agents.datagpt import LLMDataQuerier
from llmsdk.lib import SafeEncoder

from .log import *
from .lib import *

# Override default uvicorn logger...
logging.config.dictConfig(log_config)
logger = get_logger()

supported_exts = ['csv', 'sqlite']

app = FastAPI()


##############################################
# Common
##############################################
add_show_cache(app)
add_show_health(app)

##############################################
# Policy Code
##############################################
policyspec = [
    {
        "id": 1,
        "name": "filter-by-column-value",
        "description": "Allow filtering on particular column values",
        "params": [
            {
                "name": "column",
                "type": "string",
                "example": "mycolumn",
                "required": True
            },
            {
                "name": "value",
                "type": "string",
                "example": "texas",
                "required": True
            }
        ]
    }
]
add_policy(app, policyspec)

##############################################
# Handlers
##############################################
async def get_datagpt_agent(namespace, username, dataset,
                            fullpath, files,
                            metadata,
                            drop_index=False):

    cascade = get_llm_cascade()

    if 'sqlite' not in files:
        raise Exception("Please check the dataset. Only SQLite files supported for now by datagpt")

    logger.debug("Getting DataGPT agent",
                 extra={
                     "data": "Metadata: " + json.dumps(metadata, indent=4) + "\nFiles: " + str(files)
                 })

    found = False
    for filename in files['sqlite']:
        for details in metadata.get('files', []):
            if ((not isinstance(details, dict)) or
                (len(details) == 0) or
                ('path' not in details) or
                ('tables' not in details)):
                continue

            path = details['path']
            if filename.endswith(path):
                found = True
                tables = details['tables']
                print(f"Found table metadata for {filename}")

    if not found:
        raise Exception(f"Could not find tables metadata for {filename}")

    mode = 'sql'
    data = {
        'mode': mode,
        'dataset': filename,
        'tables': tables
    }

    agent = LLMDataQuerier(name="datagpt",
                           cred=None,
                           mode=mode,
                           data=data,
                           cascade=cascade,
                           debug=True)

    logger.debug(f"Choosing {mode}",
                 extra={
                     "data": json.dumps({
                     "source": "service",
                     'user': username,
                     'dataset': dataset,
                     'files': files
                     }, indent=4, cls=SafeEncoder)
                 })


    return {
        "agent": agent
    }

async def qna_run(request_id):
    """
    Query the agent instance created

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    None
         Nothing is returned. Only the result cache is updated

    """

    cache = get_cache()

    if request_id not in cache:
        logger.error(f"Failure",
                     extra={
                         'request_id': "invalid",
                         "source": "service",
                     })
        cache[request_id] = {
            'status': "failure",
            "message": f"Invalid request id"
        }
        return

    # First get the params
    value = cache[request_id]

    try:

        params  = value['params']
        user    = params['user']
        dataset = params['dataset']
        context = params['context']
        namespace = params['namespace']
        query = params['query']
        mode  = params['mode']

        stats['query_count'] += 1

        label = f"{user}_{dataset}"

        # First get the agent...
        logger.debug(f"Getting agent",
                     extra={
                         "source": "service",
                         "user": user,
                         "dataset": dataset,
                         "request_id": request_id,
                         'data': json.dumps(value, indent=4, cls=SafeEncoder)
                     })

        details = await get_agent_details(namespace,
                                          user,
                                          dataset,
                                          exts=supported_exts,
                                          get_task_specific_agent=get_datagpt_agent)
        if details is None:
            query_update_result(request_id, {
                'status': "failure",
                "message": f"LLM DataGPT Agent could not be found/built for {label}. Metadata and fileformats are typical issues. See agent log for why",
                "result": {}
            })
            return

        agent    = details['agent']
        metadata = details['metadata']

        # power/economy
        cascade_id = mode if mode in ['power', 'economy'] else None

        # Now run the query
        result = await run_in_threadpool(lambda: agent.query(query, cascade_id=mode))

        # => Answer is assumed to be a json
        result = json.loads(json.dumps(result, indent=4, cls=SafeEncoder))

        sample_response = """
        {
            "query": "how many rows are there?",
            "answer": "There are 6166 rows in the dataframe.",
            "thought": [
                ""
            ],
            "code": {
                "dialect": "pandas",
                "snippets": []
            }
        }
        """

        # Result
        result['metadata'] = metadata


        stats['query_success'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_success'] += 1
        query_update_result(request_id, {
            "status": "success",
            "result": result
        })
    except Exception as e:
        stats['query_failure'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_failure'] += 1
        logger.exception(f"Failed to run query",
                         extra={
                             "source": "service",
                             'request_id': request_id,
                             "user": params.get('user', "unknown"),
                             "dataset": params.get('dataset', "unknown"),
                         })
        query_update_result(request_id, {
            "status": "failure",
            "answer": f"Unable to construct the answer. Could be an internal agent error. See the agent log",
            "result": {
                "answer": str(e)
            }
        })



@app.get("/qna/status")
async def qna_status(request_id: str):
    """
    Get the status of the query request

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    status: dict
         Dictionary with request id, query, status, message and data
    """

    # First get the params
    try:

        cache = get_cache()

        if request_id not in cache:
            logger.error(f"Query Status: Invalid Request ID",
                         extra={
                             'request_id': request_id,
                             "source": "service",
                         })
            return {
                "request_id": request_id,
                "status": "failure",
                'answer': "Data not available. Either the request is invalid or the memory with that request has been cleared on restart"
            }

        value: dict = cache[request_id]
        status = value['status']
        logger.debug(f"Query Status: {status}",
                     extra={
                         "source": "service",
                         'request_id': request_id,
                         "dataset": value.get('dataset', 'unknown'),
                         "user": value.get('user', 'unknown'),
                         'data': json.dumps(value,indent=4, cls=SafeEncoder)
                     })

        return {
            "request_id": request_id,
            "query": value['query'],
            "status": value['status'],
            'message': value.get('message', ""),
            'data': value.get('result',{})
        }
    except Exception as e:
        return {
            "request_id": request_id,
            "query": value['query'],
            "status": "failure",
            "message": str(e),
            "data": {}
        }

@app.post("/qna")
async def qna(user: str,
              dataset: str,
              query: str,
              background_tasks: BackgroundTasks,
              mode: str = 'economy',
              context: str = "",
              namespace="datagpt",
              policy: dict = {},
              extra: dict = {}
):
    """
    Initiate a new request

    Parameters
    ----------
    user: str
          User (an unintepreted string)
    query: str
           A string
    namespace: str
           A namespace associated with this instance of the service. Multiple
    mode: str
           (economy|power) Power will result in use of GPT4
    context: str
           Any additional text
    policy: dict
           List of policies to enforce

    Returns
    -------
    status: dict
         Dictionary with request id, query, status, and data
    """

    cache = get_cache()
    request_id = str(uuid4())

    params = {
        "user": user,
        "dataset": dataset,
        "context": context,
        "namespace": namespace,
        "query": query,
        "policy": policy,
        "mode": mode
    }
    logger.debug(f"Building Index",
                extra={
                    "source": "service",
                    "user": user,
                    "dataset": dataset,
                    "data": json.dumps(params, indent=4, cls=SafeEncoder)
                })

    cache[request_id] = {
        'query': query,
        "status": "pending",
        "user": user,
        "dataset": dataset,
        "params": params
    }

    # Run the background task...
    background_tasks.add_task(qna_run, request_id)

    return {
        "request_id": request_id,
        "status": "pending",
        "query": query,
        "data": {}
    }

@app.on_event("startup")
def app_startup():

    t = BuildAgentThread(get_datagpt_agent)
    t.start()
    initialize_stats()



import os
import shutil
import logging

logger = logging.getLogger("app")


class OutputHandler(object):
    """
    How should output be handled?
    """

    def __init__(self, transform, config):
        self.transform = transform
        self.config = config

    def process(self, tempfile):
        raise Exception("Not Implemented")


class FileOutputHandler(OutputHandler):
    """
    Config in this case is a simple path.
    """

    def exists(self, params):
        output = self.config
        output = self.transform.get_file(output, abspath=False, extra=params)
        return os.path.exists(output)

    def process(self, tempfile, params):

        output = self.config
        output = self.transform.get_file(output, abspath=False, extra=params)

        # Move
        try:
            os.makedirs(os.path.dirname(output))
        except:
            pass
        shutil.move(tempfile, output)
        logger.debug(
            "Handled output",
            extra={
                "transform": self.transform.name,
                "data": "From: {}\nTo: {}".format(tempfile, output),
            },
        )


def get_mysql_config(cred):
    mapping = {
        "host": ["HOST", "host", "hostname"],
        "user": ["USER", "user", "username"],
        "port": ["port", "PORT"],
        "password": ["PASSWORD", "PASSWD", "password", "passwd", "pwd"],
        "db": ["db", "database", "NAME", "name"],
        "options": ["OPTIONS", "options"],
    }

    config = {}
    for k, alternatives in mapping.items():
        for alt in alternatives:
            if alt in cred:
                config[k] = cred[alt]
                break

    options = config.get("options", {})
    config.update(options)
    return config



"""
.. deprecated:: 2.0
   Skins were present in early versions of Enrich. Now 
   we have dashboard apps. Use them.

"""
import os
import sys
import json
import copy
import re
import traceback
import logging
import markdown
from datetime import timezone
import pytz
import humanize
import pandas as pd

from ..lib.exceptions import *
from ..lib import renderlib

logger = logging.getLogger("app")

######################################################
# Sample data
######################################################
sample_metadata = {
    "version": 1.0,
    "timestamp": "2018-03-06T19:49:42",
    "tables": {
        "voters": {
            "attributes": 22,
            "records": 3761184,
            "columns": [
                {"id": "Mobile", "label": "Mobile", "type": "text"},
                {
                    "list": [
                        {"id": "Housing", "label": "Housing", "count": 224434},
                    ]
                },
            ],
        }
    },
}

######################################################
# Global variables...
######################################################
thisdir = os.path.dirname(__file__)
templatedir = os.path.join(thisdir, "..", "templates")
templatedir = os.path.abspath(templatedir)
widgetdir = os.path.join(templatedir, "widgets")


class ExtractorSkin(object):
    def __init__(self, conf, *args, **kwargs):
        self.conf = conf
        self.required = ["extractor", "skinargs"]
        self.required_skinargs = []

    def validate_skinargs(self):

        missing = [e for v in self.required_skinargs if v not in self.skinargs]
        if len(missing) > 0:
            raise Exception("Missing variables: {}".format(missing))

    def validate_conf(self):

        missing = [
            e
            for v in self.required
            if ((v not in self.conf) and (not hasattr(self, v)))
        ]
        if len(missing) > 0:
            raise Exception("Missing variables: {}".format(missing))

        for r in self.required:
            setattr(self, r, self.conf[r])

        self.validate_skinargs()

    ####################################################
    # Rendering engine...
    ####################################################
    # Template rendering functions
    def template_render(self, widgetname, context):
        """
        Render a specified template using the context

        Args:
            widgetname (str): Name of the template
            context (dict): Key-value pairs

        Returns:
            Rendered html template that can be embedded

        """
        return renderlib.template_render(widgetdir, widgetname, context)

    def template_get_variables(self, widgetname):
        """
        Get the variables from a template

        Args:
            widgetname (str): Name of the template

        Returns:
            List of variables
        """
        return renderlib.template_get_variables(widgetdir, widgetname)

    def render_helper(self, rules, inputdata, params):
        data = []
        html = ""
        for r in rules:
            enable = r.get("enable", True)
            if not enable:
                continue
            c = r["context"](inputdata, params)
            rendered = self.template_render(r["template"], c)
            html += "\n" + rendered

        data.append({"target": "result", "rendered": html})

        return data

    def render(self, **kwargs):

        data = kwargs.pop("data")
        params = kwargs.pop("params", {})

        rules = self.skinargs["dashboard_rules"]
        return self.render_helper(rules, data, params)


class GenericExtractorSkin(ExtractorSkin):
    pass


class SearchExtractorSkin(ExtractorSkin):
    """

    This provides a search interface using the dataextractor

    """

    def __init__(self, *args, **kwargs):
        super(SearchExtractorSkin, self).__init__(*args, **kwargs)

    def validate_skinargs(self):

        super().validate_skinargs()

        extractor = self.extractor

        if not hasattr(self, "metadata"):
            self.metadata = extractor.get_metadata()

        if self.metadata is None:
            raise Exception("No metadata. So cant process")

        # Check whether all the parameters are valid
        table = self.skinargs["table"]
        if (not isinstance(table, str)) or (table not in self.metadata["tables"]):
            raise Exception("Missing or invalid table specified: {}".format(table))

        tablemetdata = self.metadata["tables"][table]
        columns = tablemetdata["columns"]

        autocomplete_column = self.skinargs["autocomplete_column"]
        columndetails = [c for c in columns if c["id"] == autocomplete_column]
        if (not isinstance(autocomplete_column, str)) or (len(columndetails) == 0):
            raise Exception(
                "Invalid or missing autocomplete column: {}".format(autocomplete_column)
            )

        if ("dashboard_rules" not in self.skinargs) or (
            not isinstance(self.skinargs["dashboard_rules"], list)
        ):
            raise Exception("Invalid dashboard rule format. Should be a list")

        invalid = []
        for i, r in enumerate(self.skinargs["dashboard_rules"]):
            if "template" not in r:
                invalid.append("Rule {} missing template".format(i))

            if ("context" not in r) or (not hasattr(r["context"], "__call__")):
                invalid.append("Rule {} invalid or missing context".format(i))

        if len(invalid) > 0:
            logger.error(
                "Invalid dashboard rules",
                extra={"transform": "SkinRenderer", "data": "\n".join(invalid)},
            )
            raise Exception("Invalid dashboard rules")

    def autocomplete(self, term, **kwargs):

        count = kwargs.pop("count", 10)
        extractor = self.extractor

        if not hasattr(self, "metadata"):
            self.metadata = extractor.get_metadata()

        if self.metadata is None:
            raise Exception("No metadata. So cant process autocomplete")

        table = self.skinargs["table"]
        autocomplete_column = self.skinargs["autocomplete_column"]

        # => Get the column details
        columns = self.metadata["tables"][table]["columns"]
        columndetails = [c for c in columns if c["id"] == autocomplete_column]
        columndetail = columndetails[0]

        if columndetail["type"] == "list":
            # Use the metadata results to search and return
            data = [
                {"label": c["label"], "key": c["id"]}
                for c in columndetail["list"]
                if (
                    (term.lower() in c["label"].lower())
                    or (term.lower() in c["id"].lower())
                )
            ]
            # Return a limited count
            data = data[:count]
            return data

        # Now this is regular 'contains' search
        # => Filter for the data
        try:

            filterargs = {
                "table": self.skinargs["table"],
                "outputs": [self.skinargs["autocomplete_column"]],
            }

            searchspec = {
                "method": "autocomplete",
                "rules": [
                    [
                        {
                            "operator": {"label": "contains", "value": "ct"},
                            "field": {
                                "label": self.skinargs["autocomplete_column"],
                                "value": self.skinargs["autocomplete_column"],
                            },
                            "value": {"label": term.lower(), "value": term.lower()},
                        },
                        {
                            "operator": {"label": "limit", "value": "limit"},
                            "field": {
                                "label": "__meta__",
                                "value": "__meta__",
                            },
                            "value": {"label": "count", "value": count},
                        },
                    ]
                ],
            }
        except:
            error = "Could not construct the search arguments"
            logger.exception(error)
            return []

        # Constructed arguments. So search
        try:
            extractor = self.extractor
            columns, records = extractor.query_by_spec(
                spec=searchspec, filterargs=filterargs
            )

            if len(records) > 0:
                data = [{"label": r[0], "key": r[0]} for r in records]
            else:
                data = []
        except:
            error = "Could not construct the dataframe required for the output"
            logger.exception(error)
            return None

        # Return a limited count
        data = data[:count]
        return data

    def search(self, query):

        # => Filter for the data
        try:
            filterargs = {"table": self.skinargs["table"], "outputs": []}

            searchspec = {
                "method": "autocomplete",
                "rules": [
                    [
                        {
                            "operator": {"label": "exact", "value": "ex"},
                            "field": {
                                "label": self.skinargs["autocomplete_column"],
                                "value": self.skinargs["autocomplete_column"],
                            },
                            "value": {"label": query, "value": query},
                        }
                    ]
                ],
            }
        except:
            error = "Could not construct the search arguments"
            logger.exception(error)
            return None

        try:
            extractor = self.extractor
            columns, records = extractor.query_by_spec(
                spec=searchspec, filterargs=filterargs
            )

            if len(records) > 0:
                # Load the records into df
                df = pd.DataFrame(records)
                df.columns = [c["name"] for c in columns]
            else:
                df = pd.DataFrame([])

        except:
            error = "Could not construct the dataframe required for the output"
            logger.exception(error)
            return None

        return df

    def render(self, *args, **kwargs):

        data = kwargs.pop("data")
        params = kwargs.pop("params", {})

        # Query the extractor
        table = self.skinargs["table"]
        autocomplete_column = self.skinargs["autocomplete_column"]

        # Now render the
        rules = self.skinargs["dashboard_rules"]
        data = self.render_helper(rules, data, params)

        return data



import os
from jinja2 import Environment, FileSystemLoader, PackageLoader, meta

thisdir = os.path.dirname(__file__)
default_templatedir = os.path.abspath(os.path.join(thisdir, "..", "templates"))

# from prompt_toolkit.token import Token
# from prompt_toolkit.styles import style_from_dict
# style = style_from_dict({
#    Token.Toolbar: '#ffffff bg:#333333',
# })


##########################
# Jinja2 helper functions...
##########################
def get_template_path(filename, templatedir=None):

    if os.path.isabs(filename):
        return filename

    if templatedir is None:
        templatedir = default_templatedir

    # Check if relative path is specified, then use it.
    # if os.path.exists(filename):
    #   return filename

    return os.path.join(templatedir, filename)


def list_templates(suffix=".node.template"):

    mapping = {}
    for t in os.listdir(default_templatedir):
        if t.endswith(suffix):
            path = os.path.join(default_templatedir, t)
            mapping[t.replace(suffix, "")] = path
    return mapping


def render(tpl_path, context):
    path, filename = os.path.split(tpl_path)
    return (
        Environment(loader=FileSystemLoader(path or "./"))
        .get_template(filename)
        .render(context)
    )


def get_variables(tpl_path):
    path, filename = os.path.split(tpl_path)
    env = Environment(loader=FileSystemLoader(path or "./"))
    template_source = env.loader.get_source(env, filename)[0]
    parsed_content = env.parse(template_source)
    return meta.find_undeclared_variables(parsed_content)


def write_rendered_file(templatefile, target, filename, params):

    if not templatefile.startswith("/"):
        templatefile = get_template_path(templatefile)

    if not os.path.exists(templatefile):
        raise Exception("Template file doesnt exist")

    content = render(templatefile, params)
    try:
        os.makedirs(target)
    except:
        pass

    with open(os.path.join(target, filename), "w") as fd:
        fd.write(content)
    return content

def get_appstore_index_data():

    currency = "inr"
    data = {
        "sidebar_targets": [
            {
                "params": { "currency": "inr"},
                "name": "inr",
                "label": "INR"
            },
            {
                "params": { "currency": "ghs"},
                "name": "ghs",
                "label": "GHS"
            },
            {
                "params": { "currency": "kes"},
                "name": "kes",
                "label": "KES"
            }
        ],
        "widgets": [
            {
                "name": f"USD-{currency.upper()} Recommendation",
                "type": "full_width_text",
                "text": "long complicated text"
            },
            {
                "name": f"USD-{currency.upper()} Dataset Details",
                "description": "Detect changes in regimes",
                "type": "full_width_key_value_pairs",
                "details": {
                    "Source": "path",
                    "Dates": 25
                }
            },
            {
                "name": f"USD-{currency.upper()} Exchange Rates",
	            "description": "Show the regimes and corresponding dates",
                "type": "full_width_image",
                "src": "Path to image or data",
                "style": "margin-left: 5px; width: 75%; height: auto;"
            },
            {
                "name": "Recommendations",
                "description": "Description of the regimes and timeframes",
                "type": "full_width_table",
                "columns": ['start_date', 'end_date', 'recommendation'],
                "rows": [
                    {
                        "start_date": o['start_date'],
                        "end_date": o['end_date'],
                        "recommendation": o['context']
                    }
                    for o in observations
                ],
                "order": [1, "dsc"],
                "style": """\
#Recommendations_table td {
    white-space: normal !important;
                }
                """
            }
        ]
    }

    return data



import re
import traceback
from texttable import Texttable

def draw_table(backend, command, data, columns, extra={}):

    t = Texttable(max_width=0)
    t.add_rows(
        [
            ["Attribute", "Value"],
            ["Name", backend.config["name"]],
            ["Server", backend.config["server"]],
            ["Command", command],
        ]
        + [[k, v] for k, v in extra.items()]
    )
    print(t.draw())

    t = Texttable(max_width=0)
    # t.set_deco(Texttable.HEADER)
    t.add_rows([columns] + [[d.get(c, "") for c in columns] for d in data])

    print(t.draw())


def draw(backend, command, data, columns, extra={}):
    return draw_table(backend, command, data, columns, extra)


def draw_dict(backend, command, data, keys=None, extra={}):

    t = Texttable(max_width=0)
    t.add_rows(
        [
            ["Attribute", "Value"],
            ["Name", backend.config["name"]],
            ["Server", backend.config["server"]],
            ["Command", command],
        ]
        + [[k, v] for k, v in extra.items()]
    )
    t.draw()

    t = Texttable(max_width=0)
    if isinstance(data, dict):
        if keys is None:
            keys = data.keys()
        t.set_deco(Texttable.HEADER)
        t.add_rows([["Key", "Value"]] + [[c, data.get(c, "")] for c in keys])
    elif isinstance(data, list):
        if keys is None:
            keys = data[0].keys()
        #t.set_deco(Texttable.HEADER)
        rows = [["Record", "Key", "Value"]]
        for idx, d in enumerate(data):
            rows.append([""]*3)
            incrows = [[str(idx), c, d.get(c, "")] for c in keys]
            rows.extend(incrows)
        t.add_rows(rows)

    print(t.draw())


def draw_run_detail(backend, command, data, extra={}):


    t = Texttable(max_width=0)
    t.add_rows(
        [
            ["Attribute", "Value"],
            ["Name", backend.config["name"]],
            ["Server", backend.config["server"]],
            ["Command", command],
        ]
        + [[k, v] for k, v in extra.items()]
    )
    print(t.draw())
    print("\n")

    if "metadata" not in data:
        return

    metadata = data["metadata"]

    print("RUN DETAILS")

    rows = [
        ["Attribute", "Value"],
        ["Start", metadata["start_time"]],
        ["End", metadata.get("end_time", "")],
        ["Command", " \n".join(metadata["cmdline"])],
        ["User", metadata["stats"]["user"]],
        ["Host", metadata["stats"]["platform"]["node"]],
    ]

    notes = metadata.get("summary", []) + metadata.get("performance_summary", [])
    if len(notes) > 0:
        rows.append(["Summary", "\n".join(notes)])

    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())
    print("\n")

    print("OUTPUTS")

    rows = [["Frame", "Type", "Description", "Size", "Path"]]
    for name, details in metadata.get("details", {}).items():
        frametype = details["frametype"]
        description = ""
        path = ""
        filesize = ""
        params = details.get("params", {})
        for p in params:
            if p.get("action", None) != "output":
                continue

            description = p.get("descriptions", p.get("description", ""))
            if isinstance(description, list):
                description = [d[:30] for d in description]
                description = "\n".join(description)
            else:
                description = description[:30]
            components = p.get("components", [])
            if len(components) > 0:
                path = "..." + components[0]["filename"][-30:]
                filesize = components[0]["filesize"]
            break
        rows.append([name, frametype, description, filesize, path])

    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())
    print("\n")

    print("VERSIONS")

    rows = [["Label", "Description", "Release", "Commit"]]
    for v in metadata["versionmap"]:
        repo = v["repo"]
        label = v["label"]
        commit = v["commit"][:8]
        description = v["description"][:20]
        release = v["release"]
        if repo.startswith("scribble-"):
            continue
        rows.append([label, description, release, commit])
    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())
    print("\n")

    if "process" in metadata:
        print("RESOURCES")
        rows = [["Attribute", "Value"]]
        for v in metadata["process"]:
            if v[0] in ["VmPeak", "VmSize", "MemTotal", "MemFree"]:
                rows.append(v)
        t = Texttable(max_width=0)
        t.add_rows(rows)
        print(t.draw())
        print("\n")

    print("LOG")
    log = data["log"]
    rows = [["Elapsed", "Transform", "Level", "Message"]]
    for l in log:
        try:
            rows.append(
                [
                    l["elapsed"],
                    l.get("transform", ""),
                    l["levelname"],
                    l["message"][:30],
                ]
            )
        except:
            traceback.print_exc()
    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())


def draw_health(backend, command, data, extra={}):

    def fixedlength(paragraph, n):
        paragraph = paragraph.strip()
        return "\n".join([paragraph[i : i + n] for i in range(0, len(paragraph), n)])

    for section in data:
        print(section["title"])
        print("-----")
        print(section["description"])
        t = Texttable(max_width=0)
        rows = [["Option", "Value", "Description"]]
        for o in section["options"]:
            value = o["value"]
            value = value.replace("<br>", "\n")
            value = value.replace("</th>\n", "</th>")
            value = value.replace("</td>\n", "</td>")
            value = re.sub("<[^>]+>", "", value)
            name = fixedlength(o["name"], 20)
            desc = fixedlength(o["description"], 20)
            rows.append([name, desc, value])
        t.add_rows(rows)
        print(t.draw())
        print("\n")



"""
Simple LLM Proxy that works with the front end. This derives from the QnA service module.

"""
import os
import sys
import re
import glob
import yaml
import json
import traceback
import time
import hashlib
import struct
import logging
from logging.config import dictConfig
from uuid import uuid4
from datetime import datetime

from fastapi import FastAPI, BackgroundTasks, Depends
from fastapi.concurrency import run_in_threadpool
from fastapi_utils.tasks import repeat_every

from llmsdk.lib import SafeEncoder
from litellm import completion

from .log import *
from .lib import *

# Override default uvicorn logger...
logging.config.dictConfig(log_config)
logger = get_logger()

app = FastAPI()

# Insert policy
policyspec = []
add_policy(app, policyspec)

"""
Samples responses of llmlite 

        sample_response = {
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "message": {
                        "content": "I'm sorry, but I don't have the capability to provide real-time weather information. However, you can easily check the weather in San Francisco by searching online or using a weather app on your phone.",
                    "role": "assistant"
                    }
                }
            ],
            "created": 1691790381,
            "id": "chatcmpl-7mUFZlOEgdohHRDx2UpYPRTejirzb",
            "model": "gpt-3.5-turbo-0613",
            "object": "chat.completion",
            "usage": {
                "completion_tokens": 41,
                "prompt_tokens": 16,
                "total_tokens": 57
            }
        }
"""

class ProxyAgent():

    def completion(self, request_id, params):

        mode = params.get("mode", "gpt-3.5-turbo")
        model = "gpt-3.5-turbo" if mode == "normal" else "gpt-4"

        query = params['query']

        completion_params = { 
            "model": model,
            "messages": [
                {
                    "content": query,
                    "role": "user"
                }
            ]
        }

        try:
            response = completion(**completion_params)
            logger.debug(f"Response from server",
                             extra={
                                 "source": "service",
                                 'request_id': request_id,
                                 "user": params.get('user', "unknown"),
                                 "dataset": params.get('dataset', "unknown"),
                                 'data': json.dumps(response, indent=4, cls=SafeEncoder)
                             })
            final_response = {
                "status": response.get('status', 'success'),
                "query": query,
                "answer": response['choices'][0]['message']['content'],
                'metadata': response
            }
            
        except Exception as e:
            logger.exception(f"Failure during execution",
                             extra={
                                 "source": "service",
                                 'request_id': request_id,
                                 "user": params.get('user', "unknown"),
                                 "dataset": params.get('dataset', "unknown"),
                         })
            final_response = {
                "status": "failure",
                "query": query,                
                "answer": "Error while accessing the LLM backend"
            }        
    
        return final_response

proxy = ProxyAgent()

@app.get("/cache")
def cache_health():
    """
    Summarize the content of the cache
    """

    cache = get_cache()
    summary = defaultdict(int)
    for k, v in cache.items():
        summary[v['status']] += 1

    return summary

@app.get("/health")
def health():
    """
    Return usage statistics

    Returns
    -------
    stats: dict
           Usage statistics

    """
    logger.info("Returning stats",
                extra={
                    'data': json.dumps(stats, indent=4)
                })
    return stats


def query_update_result(request_id, result):

    # First get the params
    cache = get_cache()
    value = cache[request_id]

    for k, v in result.items():
        value[k] = result[k]

    status = result.get('status', 'unknown')
    logger.debug(f"Result: {status}",
                extra={
                    "request_id": request_id,
                    "source": "service",
                    "user": value.get("user", "unknown"),
                    "dataset": value.get("dataset", "unknown"),
                    'data': json.dumps(result, indent=4, cls=SafeEncoder)
                })

async def query_run(request_id):
    """
    Query the agent instance created

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    None
         Nothing is returned. Only the result cache is updated

    """

    cache = get_cache()

    if request_id not in cache:
        logger.error(f"Failure",
                     extra={
                         'request_id': "invalid",
                         "source": "service",
                     })
        cache[request_id] = {
            'status': "failure",
            "message": f"Invalid request id"
        }
        return

    # First get the params
    value = cache[request_id]

    try:
        params  = value['params']
        user    = params['user']
        dataset = params['dataset']
        query   = params['query']
        namespace = params['namespace']
        label = f"{user}_{dataset}"

        logger.debug(f"{label}: Running Query",
                     extra={
                         "source": "service",
                         "user": user,
                         "dataset": dataset,
                         "request_id": request_id,
                         'data': json.dumps(value, indent=4, cls=SafeEncoder)
                     })

        stats['query_count'] += 1


        result   = await run_in_threadpool(lambda: proxy.completion(request_id, params))
        result['request_id'] = request_id

        logger.debug(f"Raw result received",
                     extra={
                         "source": "service",
                         "request_id": request_id,
                         "user": user,
                         "dataset": dataset,
                         "data": json.dumps(result, indent=4, cls=SafeEncoder)
                     })
        result['answer'] = result['answer'].strip()
        stats['query_success'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_success'] += 1
        query_update_result(request_id, {
            "status": "success",
            "result": result
        })
    except Exception as e:
        stats['query_failure'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_failure'] += 1
        logger.exception(f"Failed to run query",
                         extra={
                             "source": "service",
                             'request_id': request_id,
                             "user": params.get('user', "unknown"),
                             "dataset": params.get('dataset', "unknown"),
                         })
        query_update_result(request_id, {
            "status": "failure",
            "message":  f"Unable to construct the answer. Internal error. See the agent log",
            "result": {
                "question": query,
                "answer": "Error while accessing the LLM Agent. See log",
                "raw": [
                    str(e)
                ]
            }
        })



@app.get("/qna/status")
async def query_status(request_id: str):
    """
    Get the status of the query request

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    status: dict
         Dictionary with request id, status, message and data

    """
    # First get the params
    try:

        value: dict = {}

        cache = get_cache()

        if request_id not in cache:
            logger.error(f"Query Status: Invalid Request ID",
                         extra={
                             'request_id': request_id,
                             "source": "service",
                         })
            return {
                "request_id": request_id,
                "status": "failure",
                "message": "Request ID not found"
            }

        value = cache[request_id]
        status = value['status']
        response = {
            "request_id": request_id,
            "status": value['status'],
            'message': value.get('message', ""),
            'data': value.get('result',{})
        }
        logger.debug(f"Query Status: {status}",
                     extra={
                         "source": "service",
                         'request_id': request_id,
                         "dataset": value.get('dataset', 'unknown'),
                         "user": value.get('user', 'unknown'),
                         'data': json.dumps(response,indent=4, cls=SafeEncoder)
                     })
        return response

    except Exception as e:
        logger.exception("Query Status: failure",
                     extra={
                         "source": "service",
                         'request_id': request_id,
                         "user": value.get('user',"unknown"),
                         "dataset": value.get('dataset',"unknown"),
                     })

        return {
            "request_id": request_id,
            "status": "failure",
            "message": str(e),
            "data": {}
        }

@app.post("/qna")
async def query_index(user: str,
                      dataset: str,
                      query: str,
                      background_tasks: BackgroundTasks,
                      mode: str = 'internal',
                      context: str = "",
                      namespace="default",
                      policy: dict = {},
                      extra: dict = {}
):
    """
    Initiate a new request

    Parameters
    ----------
    user: str
          User (an unintepreted string)
    query: str
           A string
    namespace: str
           A namespace associated with this instance of the service. Multiple
    context: str
           Any additional text
    policy: dict
           List of policies to enforce
    extra: dict
           Pass any other information

    Returns
    -------
    status: dict
         Dictionary with request id, query, status and data
    """

    cache = get_cache()
    request_id = str(uuid4())

    params = {
        "user": user,
        "dataset": dataset,
        "query": query,
        "mode": mode,
        "context": context,
        "namespace": namespace,
        "policy": policy
    }
    label = f"{user}_{dataset}"

    if label not in stats['datasets']:
        stats['datasets'][label] = {
            "query_count": 0,
            "query_success": 0,
            "query_failure": 0,
        }
        
    
    logger.debug(f"Starting query",
                extra={
                    "source": "service",
                    "user": user,
                    "dataset": dataset,
                    "data": json.dumps(params, indent=4, cls=SafeEncoder)
                })

    cache[request_id] = {
        "status": "pending",
        "user": user,
        "dataset": dataset,
        "query": query,        
        "params": params
    }

    # Run the background task...
    background_tasks.add_task(query_run, request_id)

    return {
        "request_id": request_id,
        "status": "pending",
        "data": {}
    }

@app.on_event("startup")
def app_startup():

    initialize_stats()



"""
File contains different classes for time series forecasting

Classes:
BaseProphetForecasterModel - Class that uses prophet library to forecast 

"""


import math
import ruptures as rpt
from collections import defaultdict
import matplotlib.pyplot as plt
from prophet import Prophet
from prophet.plot import add_changepoints_to_plot

__all__ = ['BaseProphetForecasterModel']


class BaseProphetForecasterModel(object):
    """
    Base class for time series forecasting
    """

    def __init__(self, df, *args, **kwargs):
        """
        defaults
        """
        self.data = df

        self.growth =  'linear'
        self.horizon = 3
        self.periods = self.horizon * 2
        self.seasonality_mode = 'multiplicative'

        super().__init__(*args, **kwargs)

    
    def _init_model(self, growth, seasonality_mode):

        model = Prophet(growth=growth,
                        changepoint_prior_scale=0.5,
                        seasonality_mode=seasonality_mode)
        self.model = model



    def run_forecasting(self, params):
        # setup and run

        df = self.data

        holidays = params.get('holidays', [])
        regressors = params.get('regressors', [])
        growth = params.get('growth', self.growth)
        periods = params.get('periods', self.periods)
        seasonality_mode = params.get('seasonality_mode', self.seasonality_mode)

        self._init_model(growth, seasonality_mode)
        model = self.model

        # add country holidays
        for country in holidays:
            model.add_country_holidays(country_name=country)
        
        # add any additional regressors
        for r in regressors:
            model.add_regressor(r)
        
        # set the cap for the dataframe
        # needed for logitstic model 
        cap = df['y'].max()
        if growth == 'logistic':
            df['cap'] = cap
        
        # fit the model 
        model.fit(df)
        
        # create a future timeseries set for N periods
        future = model.make_future_dataframe(periods=periods)
        # set the cap
        if growth == 'logistic':
            future['cap'] = cap

        # add any additional regressors to the future timeseries
        for r in regressors:
            future[r] = df[r].tail(1).values[0]

        # make predictions using the fitted model
        forecast = model.predict(future)
        
        return forecast


    def visualize_forecasting(self, forecast, chart_params):
        """
        visualize the changepoints
        """
        model = self.model

        xlabel = chart_params.get('xlabel', 'Date')
        ylabel = chart_params.get('ylabel', 'Value')
        title = chart_params.get('title', 'Forecast')
        show_changepoints = chart_params.get('show_changepoints', False)

        fig = model.plot(forecast)

        fig.gca().set_xlabel(xlabel)
        fig.gca().set_ylabel(ylabel)
        fig.gca().set_title(title)

        if show_changepoints:
             add_changepoints_to_plot(fig.gca(), model, forecast)

        return fig



import os
import sys
import json
import glob
import time
import copy
import traceback
import tempfile
import shutil
import numpy as np
import pandas as pd
from enrichsdk import Compute, S3Mixin
from datetime import datetime, timedelta, date
import logging
import s3fs
from dateutil import parser as dateparser

from enrichsdk.contrib.lib.transforms import MetricsBase
from enrichsdk.lib.integration import send_html_email
from enrichsdk.utils import (note, get_today, get_yesterday, get_tomorrow)

# Define a registry
from libtip.datasets import get_registry

logger = logging.getLogger("app")
thisdir = os.path.dirname(__file__)

from .profilespec import get_profile_spec

class My{{transform_name}}(S3Mixin, MetricsBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "{{transform_name}}"
        self.description = "{{transform_description}}"
        self.author = "{{author_name}}"

        self.supported_extra_args = [
            {
                "name": "run_date",
                "description": "Date of execution",
                "default": get_today,
                "required": False,
            },
            {
                "name": "specs",
                "description": "specs to run (all|<add list>)",
                "default": "all",
                "required": False,
            },
            {
                "name": "send_email",
                "description": "Send daily notification email",
                "default": "false",
                "required": False,
            },
        ]

        test_root = os.environ['ENRICH_TEST']
        self.testdata = {
            'data_root': os.path.join(test_root, self.name),
            'inputdir': test_root,
            'outputdir': os.path.join(test_root, self.name, 'output'),
            'statedir': os.path.join(test_root, self.name, 'state'),
            'conf': {
                'args': {
                    "run_date": get_today(),
                    "s3cred": "demouser",
                    's3root':   "enrich-demodata/backup/%(node)s/data/acme/Core/shared/emetrics",
                    "specs": "all",
                    "receivers": {}
                }
            },
            'data': {
            }
        }

    @classmethod
    def instantiable(cls):
        return True

    def preload_clean_args(self, args):

        args = super().preload_clean_args(args)

        args['s3root'] = self.get_file(args['s3root'], abspath=False)
        args['s3cred'] = s3cred = self.get_credentials(args['s3cred'])
        args['s3'] = s3fs.S3FileSystem(anon=False,
                                       key=s3cred['access_key'],
                                       secret=s3cred['secret_key'])

        # Insert these variables...
        run_date = dateparser.parse(args['run_date']).date()
        args['run_date'] = run_date
        args['end_date'] = args['run_date'] = run_date
        args['start_date'] = args['end_date'] + timedelta(days=-31)

        return args

    #################################
    # Get data
    #################################
    def get_db_uri(self, source):
        uri = source['uri']
        uri = self.get_file(uri, abspath=False)
        return uri

    def get_profile(self):
        """
        Construct ore turn profile.
        """
        return get_profile_spec()

    def get_specs(self, profile):
        """
        Return a list of valid specifications...
        """
        # Get all available specs
        specs = super().get_specs(profile)

        # now filter them out...
        if 'all' in self.args['specs']:
            return specs

        selected = [s for s in specs if s['name'] in self.args['specs']]
        return selected

    def get_dataset(self, name):
        """
        Get dataset by
        """
        registry = get_registry(transform=self, state=self.state)
        datasetobj = registry.find(name)
        if datasetobj is None:
            raise Exception(f"Unknown dataset: {name}")
        return datasetobj

    def read_data(self, filename, params, **kwargs):

        s3 = self.args['s3']
        if s3.exists(filename):
            df = pd.read_csv(s3.open(filename),**params)

            if (('ledger_balance' in filename) or
                ('ledgers_cash_flow_analysis' in filename) or
                ('mtn' in filename)
            ):
                df['snapshot_date'] = kwargs['date']

            return df

        return None

    ##########################
    # State management
    ##########################
    def update_state(self, spec, framename, framedesc, framedf, s3path):
        """
        Save the results for profiling..
        """

        s3 = self.args['s3']

        state = self.state
        framemgr = self.config.get_dataframe('pandas')

        # Generate column information...
        params = self.get_column_params(framename, framedf)

        params.append({
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "dataframe",
                        "nature": "input",
                        "objects": spec['sources']
                    },
                    {
                        "type": "file",
                        "nature": "output",
                        "objects": [s3path]
                    }
                ]
        })

        ## => Gather the update parameters
        updated_detail = {
            'df': framedf,
            'description': framedesc,
            'transform': self.name,
            'frametype': 'pandas',
            'params': params,
            'history': [
                # Add a log entry describing the change
                {
                    'transform': self.name,
                    'log': 'Generated the dataset'
                }
            ]
        }

        # Dump it into the shared state
        state.update_frame(framename, updated_detail, create=True)

    ##############################
    # Store in S3/Other...
    ##############################
    def store_common(self, data, spec, descriptions={}):
        """
        Store in s3. This code assumes that s3 handle is
        available. Replace as required.
        """

        s3root = self.args['s3root']
        run_date = self.args['run_date']
        version = "v1"

        # Replace as required...
        s3   = self.args['s3']

        # Pipeline metadata
        metadata = self.get_default_metadata(self.state)

        # Insert column information
        columns = {}
        for name, df in data.items():
            extra = self.get_column_metadata(name, df)
            columns[name]  = extra
        metadata['columns'] = columns

        root = os.path.join(s3root, version, run_date.isoformat())
        metadatafile = os.path.join(root, f"{spec['name']}.metadata.json")

        # Now dump the data
        with s3.open(metadatafile, 'w') as fd:
            fd.write(json.dumps(metadata, indent=4))
        msg = "metadata: {}\n".format(metadatafile)

        # => Destination details...
        for name, df in data.items():
            filename = os.path.join(root, f"{spec['name']}.{name}.csv")
            msg += f"{name}: {filename}\n"
            with s3.open(filename, 'w') as fd:
                df.to_csv(fd, index=False)

            # Update the state with the dataframe...
            self.update_state(spec,
                              f"{spec['name']}-{name}",
                              descriptions.get(name,name),
                              df, filename)

        msg += '\n'
        logger.debug(f"{spec['name']} wrote to s3: {run_date.isoformat()}",
                     extra={
                         'transform': self.name,
                         'data': msg
                     })

    def send_common(self, spec, attachments, note=None):
        """
        Send email/other notifications. with required
        attachments. This is useful when the metrics needs to be
        'pushed'
        """

        run_date = self.args['run_date']
        allreceivers = self.args['receivers'].get('all', [])
        receivers = self.args['receivers'].get(spec['name'], allreceivers)
        if len(receivers) == 0:
            logger.debug(f"Skipping notifications. No receivers",
                         extra={
                             'transform': self.name,
                         })
            return

        html = f"""\
        <h2>{spec['description']} - {run_date.isoformat()}</h2>

        """
        if note is not None:
            html += note

        # Replace as required...
        sender='Analytics Team<support@scribbledata.io>'
        subject=f"{spec['description']} - {format(run_date.isoformat())}"
        send_html_email(html,
                        sender=sender,
                        receivers=receivers,
                        subject=subject,
                        attachments=attachments)

        logger.debug("Email notification sent",
                     extra={
                         'transform': self.name,
                         'data': "Receivers: " + json.dumps(receivers)
                     })

    def clean_query_result(self, df, source):

        name = source['name']
        logger.debug(f"Cleaning {name}")

        aggspec = {
            'total_txn_count': 'sum',
        }
        for col in df.columns:
            if col.endswith("_unique") or col in ['months', 'nationality_cleaned']:
                aggspec[col] = lambda s: ";".join(s.unique())

        df1 = df.groupby('unique_key').agg(aggspec)

        # Label customer as repeat customer
        df1['repeat'] = df1['total_txn_count'].map(lambda x: 1 if x > 1 else 0)

        df1 = df1.reset_index()

        return {
            "query_data_cleaned": df1
        }

    def generate_result(self, data, spec):

        df = data[spec['name']]

        # Do something complicated...
        df = df.rename(columns={
            'lifetime_mean': "lifetime_mean_days",
        })

        df = df.sort_values(['dimensions','all_customers'],  ascending=[True, False])

        # Return a set of processed dataframes
        return {
            spec['name']: df
        }

    ###########################################
    # Store in multiple forms...
    ###########################################
    def store_results_excel(self, dirname, data, spec):
        """
        Override as required..
        """
        run_date = self.args['run_date']

        filename = os.path.join(dirname, f'{spec["name"]}-{run_date.isoformat()}.xlsx')
        writer = pd.ExcelWriter(filename,engine='xlsxwriter')
        workbook=writer.book

        # Write some or all the dataframes in the data dictionary
        for name, df in data.items():
            df.to_excel(writer, sheet_name=name, startrow=0, startcol=0, index=False)

        # Write the sheets
        writer.save()

        return filename

    def store_result(self, spec, data):
        """

        """
        run_date = self.args['run_date']

        self.store_common(data, spec)

        # Should I send an email/notification:
        if not self.args['send_email']:
            logger.debug(f"Skipping sending email",
                         extra={
                             'transform': self.name
                         })
            return

        note="<p>Note: Any additional notes to add</p>"

        try:
            tmpdir = tempfile.mkdtemp()
            filename = self.store_results_excel(tmpdir, data, spec)
            self.send_common(spec, note=note, attachments=[filename])


        except:
            logger.exception("Unable to send email",
                         extra={
                             'transform': self.name,
                         })

        if os.path.exists(tmpdir):
            try:
                shutil.rmtree(tmpdir)
            except:
                logger.exception("Unable to cleanup temporary directory",
                                 extra={
                                     'transform': self.name,
                                     'data': tmpdir
                                 })


provider = My{{transform_name}}




import os
import sys
import json
import re
import numpy as np
from sqllineage.runner import LineageRunner
from datetime import date, datetime

from .sample import *
from .excel import *
def get_lineage_of_query(engine, sql):
    """
    Given an engine and an SQL expression, find the input
    and output tables. Replace <default> with database
    name from the engine
    """
    dependencies = []
    result = LineageRunner(sql, verbose=True)

    database = engine.url.database
    if database is None:
        database = "<default>"

    # Get input tables..
    if len(result.source_tables) > 0:
        tables = [str(t).replace("<default>", database) for t in result.source_tables]

        tables = {
            t.split(".")[-1]: {"text": t.replace(".", " "), "table": t} for t in tables
        }
        dependencies.append(
            {
                "type": "db",
                "nature": "input",
                "objects": tables,
            }
        )

    # Get output dependencies
    if len(result.target_tables) > 0:
        tables = [str(t).replace("<default>", database) for t in result.target_tables]
        tables = {
            t.split(".")[-1]: {"text": t.replace(".", " "), "table": t} for t in tables
        }

        dependencies.append({"type": "db", "nature": "output", "objects": tables})

    return dependencies

class SafeEncoder(json.JSONEncoder):

    def default(self, obj):
        try:
            if isinstance(obj, (datetime, date)):
                result = obj.isoformat()
            elif isinstance(obj, (tuple)):
                result = super().default(list(obj))
            elif isinstance(obj, np.integer):
                return super().default(int(obj))
            elif isinstance(obj, float) and np.isnan(obj):
                return "null"
            elif isinstance(obj, np.floating):
                return float(obj)
            elif isinstance(obj, np.ndarray):
                return obj.tolist()
            else:
                result = super().default(obj)
        except:
            result = str(obj)

        return result

def clean_nans(data):

    if (isinstance(data, float) and
        (np.isnan(data) or np.isinf(data))):
        return "null"

    if isinstance(data, (list)):
        return [clean_nans(d) for d in data]
    elif isinstance(data, (tuple)):
        return (clean_nans(d) for d in data)
    elif isinstance(data, dict):
        return {
            clean_nans(k): clean_nans(v)
            for k, v in data.items()
        }

    return data

def make_safely_encodable(data):
    """
    Make sure that the data is safely processable by modules
    """
    data = clean_nans(data)
    data = json.loads(json.dumps(data, allow_nan=False, cls=SafeEncoder))
    return data

##################################
# Datetime helpers
##################################
from datetime import datetime, date, timedelta

def get_today():
    return date.today().isoformat()

def get_yesterday():
    yesterday = date.today() + timedelta(days=-1)
    return yesterday.isoformat()

def get_daybefore():
    daybefore = date.today() + timedelta(days=-2)
    return daybefore.isoformat()

def get_lastweek():
    lastweek = date.today() + timedelta(days=-7)
    return lastweek.isoformat()

##################################
# Helper...
##################################
def note(df, title):
    """
     Quick summary of a dataframe including shape, column, sample etc.

     Args:
        df (dataframe): Input dataframe
        title (str): Title

    Returns:
        str: A formatted text to be used for logging

    """

    # May be the parameters have been flipped
    if isinstance(df, str):
        df, title = title, df

    msg = title + "\n"
    msg += "--------" + "\n"
    msg += "Timestamp: " + str(datetime.now()) + "\n"
    msg += "\nShape: " + str(df.shape) + "\n"
    msg += "\nColumns: " + ", ".join([str(c) for c in df.columns]) + "\n"
    if len(df) > 0:
        msg += "\nSample:" + "\n"
        msg += df.sample(min(2, len(df))).T.to_string() + "\n" + "\n"
    msg += "\nDtypes" + "\n"
    msg += df.dtypes.to_string() + "\n"
    msg += "------" + "\n"
    return msg

############################

def get_month():
    return datetime.now().strftime("%Y-%m")

def get_yesterday():
    yesterday = date.today() + timedelta(days=-1)
    return yesterday.isoformat()

def get_tomorrow():
    tomorrow = date.today() + timedelta(days=1)
    return tomorrow.isoformat()

def get_today():
    return date.today().isoformat()

def get_daybefore():
    daybefore = date.today() + timedelta(days=-2)
    return daybefore.isoformat()

########################################
# Text cleaner...
########################################
def slugify(s):
  s = s.lower().strip()
  s = re.sub(r'[^\w\s-]', '', s)
  s = re.sub(r'[\s_-]+', '_', s)
  s = re.sub(r'^-+|-+$', '', s)
  return s



import os
import json
import time
import string
import hashlib

from langchain.chat_models import ChatOpenAI
from langchain.schema import AIMessage, HumanMessage, SystemMessage
from langchain.callbacks import get_openai_callback

from . import agent_events
from llmsdk.lib import SafeEncoder
from llmsdk.services.log import *

__all__ = ['LLMMultiLabelTextClassifier']

class LLMMultiLabelTextClassifier(object):
    """
    A bot to take in some text and perform multi-label classification on it
    """

    def __init__(self,
                 name,
                 cred={},
                 platform="openai"):
        """
        init the bot
        name: name of the bot
        cred: credentials object
        platform: name of the platform backend to use
                default to openai platform for now
                will be extended in the future to suuport other platforms
        """

        start_time = time.time()

        # logging
        self.logger = get_logger()

        # defaults
        self.metadata = {}
        self.sanity_entries = {}
        self.max_llm_tokens = 512

        # name
        self.agent_name = name
        self.agent_type = "query-router"

        # creds
        self.cred = cred
        # LLM params
        self.platform = platform

        # init the llm and embeddings objects
        self.llm = self._get_llm_objs(platform=self.platform,
                                      cred=self.cred)

        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "platform": self.platform,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## helper functions

    def _create_id(self, string):
        """
        create a unique identifier from a string
        """
        return hashlib.md5(string.encode('utf-8')).hexdigest()

    def _log_event(self, event_name, duration=0, params={}):
        """
        format an event object to store in the agent's metadata
        event_name: name of the event to log
        duration: how long did this event take, only needed for events
                   that require some waiting (e.g. RTT to LLM APIs)
        params: any other info needed to be logged
        """

        ts = time.time() - duration # to get the actual start time
        event_id = self._create_id(f'{event_name}-{ts}')

        event = {
            "id": event_id,
            "agent": self.agent_name,
            "agent_type": self.agent_type,
            "timestamp": round(ts, 3),
            "duration": round(duration, 3),
            "name": event_name,
            "params": params,
        }
        self.metadata['events'].append(event)

        return event

    def _get_llm_objs(self, platform, cred):
        # get the api key from creds
        api_key = self._get_api_key(cred, platform)

        # init the model
        if platform == "openai":
            # get the llm object
            llm = ChatOpenAI(temperature=0,
                             openai_api_key=api_key,
                             model="gpt-3.5-turbo",
                             max_tokens=self.max_llm_tokens,
                             request_timeout=60)
        else:
            llm = None

        return llm

    def _get_api_key(self, cred, key):
        """
        get the API key from the cred
        """
        api_key = None
        if isinstance(cred, str):
            api_key = cred
        if isinstance(cred, dict) and key in cred:
            api_key = cred[key]
        return api_key

    def _err_msg(self, t):
        msgs = {
            "field": "I'm having trouble understanding. Try another way of wording your query."
        }
        return msgs.get(t, "Something went wrong, try your query again.")

    ## interfaces

    def get_metadata(self):
        """
        return metadata collected by the agent
        """
        return self.metadata

    def get_prompt(self, query, label_spec):
        """
        generate a prompt for querying the LLM
        """
        # construct the prompt template

        defualt_persona = """You are a highly advanced, AI-enabled bot, that performs multi-label text classification"""

        # get the bot's persona
        persona = label_spec.get("persona", defualt_persona)

        # get the bot's specified instruction set
        instructions = label_spec.get("instructions", "")

        # get the synonymns
        syns = []
        for term, synonymns in label_spec.get('synonymns', {}).items():
            syn = f"  - {term}: {', '.join(synonymns)}"
            syns.append(syn)
        synonymns = "\n".join(syns)
        if len(synonymns) > 0:
            synonymns = f"""Here is a set of comma-separated synonymns for various phrases that may be asked about:
{synonymns}"""

        # get the output format
        output_format = label_spec.get('output_format', {})
        output_type = output_format.get('type', 'json')
        output_sample = output_format.get('sample', None)

        if output_sample:
            output_format = f"""Always respond by formatting your response as a {output_type} object EXACTLY as follows:
{output_sample}"""
        else:
            output_format = f"""Always respond by formatting your response as a {output_type} object"""

        # construct the system message
        sys_msg = "\n\n".join([persona, instructions, synonymns, output_format])

        # construct the human message
        human_msg = f"""
Here is the question:
------ BEGIN QUESTION ------
{query}
------- END QUESTION -------

Your response:
"""

        messages = [
            SystemMessage(content=sys_msg),
            HumanMessage(content=human_msg),
        ]

        return messages

    def label(self, query, label_spec={}):
        """
        run a prompt against the LLM using the routing spec
        """
        start_time = time.time()

        success = True
        response = None

        # construct the prompt to the policy bot
        prompt = self.get_prompt(query, label_spec)

        # run the query
        try:
            if self.platform in ['openai', 'azure']:
                with get_openai_callback() as cb:
                    response = self.llm(prompt)
                stats = {
                    "total_tokens": cb.total_tokens,
                    "prompt_tokens": cb.prompt_tokens,
                    "completion_tokens": cb.completion_tokens,
                    "total_cost": round(cb.total_cost, 4)
                }
            else:
                response = self.llm(prompt)
                stats = {}

            response = json.loads(response.content)

        except:
            success = False

        # log the event
        params = {
            "query": query,
            "mode": "internal",
            "result": response.copy() if response is not None else None,
            "stats": stats,
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        return success, response


def get_sample_profilespec():

    default = {

        "persona": """You are a highly advanced, AI-enabled bot whose job is to take a question from a user and assign two labels to it.
The first label is what kind of insight is being asked about.
The second label is what kind of cohort is being asked about.""",

        "instructions": """The questions will be about the results of a survey across multiple respondent cohorts for multiple products.
Note that sometimes a question can be about all the products in the survey without asking about population cohorts, and sometimes cohort information will be asked.
The three types of population cohorts are: region, gender, age

The following are metrics that could be mentioned in the question:
  - 'reach' a.k.a. 'line insights' a.k.a. 'le'
  - 'voc' a.k.a. 'sentiment score' a.k.a. 'ss'

Here are the rules by which you must perform the insight label assignment:
  - If the question is asking about actionable insights, then assign insight=actionable_insights
  - If the question is asking specifically about best performing or worst performing products, then assign insight=actionable_insights
  - If the question is asking about product metrics with a specific cohort, then assign insight=le_insights
  - If the question is asking about the affinity of a population cohort to products, then assign insight=ss_insights
  - If the question is asking about product features, then assign insight=features_insights
  - If the question is asking about the audience, then assign insight=audience_insights

And here are the rules by which you must perform the cohort label assignment:
  - If the question is asking about groups of respondents, then assign cohort=audience
  - If the question is asking about locations, then assign cohort=region
  - If the question is asking about gender, then assign cohort=gender
  - If the question is asking about age, then assign cohort=age
  - If the question is asking about best, or winners, or hits then assign cohort=best
  - If the question is asking about worst or losers then assign cohort=worst
  - If the question is asking about not liking something, then assign cohort=dislikes
  - If the question is asking about one metric being high and another being low, then assign cohort=mixed""",

        "synonymns": {
            "actionable insights": [
                "survey summary",
                "global summary",
                "overall performance"
            ],
            "line efficiency": [
                "le",
                "reach",
                "cannibalization",
                "incremental reach",
                "cumulative reach",
            ],
            "sentiment score": [
                "ss",
                "voc",
                "voice of consumer",
                "sentiment graph",
                "consumer graph",
                "consumer appetite",
                "consumer sentiment",
            ],
        },

        "output_format": {
            "type": "json",
            "sample": '{"insight": insight, "cohort": cohort}'
        }
    }

    return default


if __name__ == "__main__":

    def query_to_text(agent, query, label_spec, commentary):
        """
        Take a query and a label_spec and use the MLTClassifier agent to get a set of labels
        then use the labels to construct a key
        use the key to lookup the data in the commentary dict
        """
        # default
        err = "Error"
        text = err

        # first, get the route
        success, result = agent.label(query, label_spec)

        if success:
            # construct the key
            keyparts = []
            keyparts.append(result.get('insight', ''))
            keyparts.append(result.get('cohort', ''))
            keyparts = [k for k in keyparts if k!='']
            key = "_".join(keyparts)

            # lookup
            text = commentary.get(key, {}).get('text', err)

        return text

    # vars
    commentaryfile = "..." # /path/to/{survey}-output.json
    agentname = "surveybot"
    platform = "openai"
    labelspec = get_sample_profilespec()

    # load the commentary data
    with open(commentaryfile, "r") as fd:
        commentary_json = json.load(fd)
    commentary = commentary_json.get('commentary', {})

    # create an agent
    surveybot = LLMMultiLabelTextClassifier(name=agentname, platform=platform)

    # query the agent
    query = "How are these products reaching with women?"
    text = query_to_text(surveybot, query, labelspec, commentary)
    print (f"Question: {query}")
    print (f"Answer: {text}")



import json
import time
import string
from re import sub

from langchain import PromptTemplate
from langchain.chains.question_answering import load_qa_chain
from langchain.callbacks import get_openai_callback

from . import agent_events
from llmsdk.agents.basellm import BaseLLMAgent
from ..lib import extractors
from llmsdk.lib import SafeEncoder

__all__ = ['LLMQuerierExtractor']

class LLMQuerierExtractor(BaseLLMAgent):
    """
    Class to do querying of a docset and extracting specific information fields using LLMs
    Query can be run against a specified set of documents that act
    as context to constrain the answers
    """

    def __init__(self,
                 name,
                 cred={},
                 platform="openai",
                 searchapi="serpapi",
                 statestore="redis",
                 topk=7):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        platform: name of the LLM platform backend to use
                default to OpenAI GPT platform for now, Azure is also supported
                will be extended in the future to suuport other models
        memory_size: how many tokens of memory to use when chatting with the LLM
        """

        start_time = time.time()

        # init the base class
        super().__init__(name=name,
                         cred=cred,
                         platform=platform,
                         agent_type="extract",
                         searchapi=searchapi,
                         statestore=statestore)

        # defaults
        self.chunk_size = 1000
        self.chunk_overlap = 300
        self.index = None
        self.metadata = {}
        self.vdb_client = None
        self.index_name = None
        self.index_store = None
        self.topk = topk
        self.doc_signatures = []
        self.docs = {}

        # LLM params
        self.platform = platform
        self.chaintype = "stuff"
        self.searchapi = searchapi

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                       cred=self.cred)

        # init the QnA chain for internal queries
        prompt = self._get_query_prompt_internal()
        self.llm_chain_int = load_qa_chain(llm=self.llm,
                                           chain_type=self.chaintype,
                                           prompt=prompt)
        # init the agent for searches
        self.llm_agent_srch, self.searchengine = self._load_search_agent(cred=self.cred,
                                                                          searchapi=self.searchapi,
                                                                          llm=self.llm)
        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "name": self.agent_type,
                "platform": self.platform,
                "chaintype": self.chaintype,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## helper functions

    def _get_query_prompt_internal(self):
        """
        generate a prompt for running a query in internal mode
        """
        template = """You are a highly advanced AI program designed to extract specific pieces of information from business documents.
Be precise in the responses you provide. Do not respond in full sentences, but only with the specific information asked for.
You will be provided with relevant pieces of context extracted from a business document to answer the question at the end.
If the question asks you to format your answer in some specific manner, do so.
If the question includes instructions to help you with extraction, follow those instructions.
If you cannot find the answer in the context, just say 'unknown', don't try to make up an answer and do not provide any explanations.

------ BEGIN CONTEXT ------
{context}
------ END CONTEXT ------

------ BEGIN QUESTION ------
{input} {instructions}
------ END QUESTION ------

Your response:"""

        prompt = PromptTemplate(
            input_variables=["input", "context", "instructions"],
            template=template
        )

        return prompt

    ## interfaces


    def read_document(self, source, content, metadata={}, params={}, store="chroma", persist_directory=None):
        """
        wrapper function that takes in the path to a document
        and sets it up for reading by the agent
        this function will create a new index if the agent does not already have one
        else it will use the existing index pointer
        needs the persist_directory that the index will use
        """
        # load the document
        data = self.load_data(source=source, content=content)

        # add the document to index
        if not self.index:
            # we have to init a new index
            self.create_add_index(data=data,
                                   store=store,
                                   persist_directory=persist_directory,
                                   index_name=self.agent_name)
        else:
            # we can use the agent's index pointer
            self.add_to_index(data)

        # extract text from document if it is a pdf
        # so that we have the table data
        if source in ["pdf"]:

            # run through Textract
            extracted_data = extract_text_from_file(content, provider="aws")

            # take the Textract output
            # and add tables and linetext to index
            for block in ["tables", "text"]:
                for extract in extracted_data:
                    # for each page in the document
                    for entry in extract[block]:
                        # for each table in the page
                        if any(f not in entry for f in ['id', 'content']):
                            continue
                        metadata = { "source": entry['id'] }
                        data = self.load_data(source="str",
                                               content=entry['content'],
                                               metadata=metadata)
                        self.add_to_index(data)

            # add the signature details to the agent's knowledge
            for extract in extracted_data:
                signatures = extract["signatures"]
                self.doc_signatures.extend(signatures)

        return

    def run_query_search(self, query, instructions=""):
        """
        run a query using the search agent
        this is useful when looking for answers using a search engine
        """
        def extract_content_sources(sourcedata):
            docs = sourcedata.get('organic_results')
            if not docs:
                return None
            sources = [{"content": d.get('snippet', ""), "source": d.get('link')} for d in docs]
            return sources

        # get the human-readable result
        result = self.llm_agent_srch.run(input=query)

        # get the sources
        sourcedata = self.searchengine.results(query)
        sources = extract_content_sources(sourcedata)
        if not sources:
            sources = [{"content": "", "source": f"search-{self.searchapi}"}]

        # construct result
        result = {
            "question": query,
            "answer": result,
            "suggest": list(set([q.get('question', '') for q in sourcedata.get('related_questions', [])])),
            "sources": sources
        }

        return result

    def run_query_internal(self, query, instructions=""):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers using a private source of data
        """
        # get the similar docs
        docs = self.get_similar_docs(query, topk=self.topk)

        # setup the QnA chain object
        response = self.llm_chain_int({"input_documents":docs, "input":query, "instructions":instructions},
                                      return_only_outputs=False)

        # run the query against the similar docs
        result = {
            "question": query,
            "answer": response.get('output_text', self._err_msg('field')).strip(),
            "sources": [{"content": d.page_content, "metadata": d.metadata, "distance": d.metadata.pop('distance')} for d in docs],
        }

        return result

    def query(self, query, instructions="", mode="internal"):
        """
        run a query on an index using an llm chain object
        query: query string
        mode: 'internal' for querying over docset, 'search' for searching the web
        """

        start_time = time.time()

        method = getattr(self, f"run_query_{mode}", None)
        if method is None:
            raise Exception(f"Unsupported mode: {mode}")

        try:
            if self.platform in ['openai', 'azure']:
                with get_openai_callback() as cb:
                    result = method(query, instructions)
                stats = {
                    "total_tokens": cb.total_tokens,
                    "prompt_tokens": cb.prompt_tokens,
                    "completion_tokens": cb.completion_tokens,
                    "total_cost": round(cb.total_cost, 4)
                }
            else:
                result = method(query, instructions)
                stats = {}
        except:
            result = {
                "question": query,
                "answer": self._err_msg('field'),
                "sources": [],
            }
            stats = {}

        # log the event
        params = {
            "query": query,
            "mode": mode,
            "result": result.copy() if result is not None else None,
            "stats": stats,
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # add the event to the result
        result['metadata'] = {
            "timestamp": event['timestamp'],
            "duration": event['duration'],
        }

        return result

    def process_spec_queries(self, spec):
        """
        take a spec containing questions and answer them
        against the docset indexed by the agent
        """

        # retry answers
        retry_answers = ["im having trouble understanding try another way of wording your query",
                         "unknown"]

        # get the list of queries
        query_set = spec.get("query_set", [])
        indx_query_set = {}

        # begin an empty dict
        extracted_info = {}
        grounding = {}

        # foreach query to process
        for one_query in query_set:

            enable = one_query.get("enable", True)
            q_name = one_query['name']
            query = one_query.get("query")
            query_mod = one_query.get("query_mod")
            query_alts = one_query.get("query_alts", [])
            use_alts = one_query.get("use_alts", "on-fail")
            postprocess = one_query.get("postprocess", {})
            fill_columns = one_query.get("fill_columns", [])
            query_mode = one_query.get("mode", "internal")
            instructions = one_query.get("instructions", "")

            # post-processing handling
            pp_handler = postprocess.get("handler")
            pp_response = postprocess.get("response", "fill")
            pp_othercols = postprocess.get("othercols", [])

            # make note of this one_query in the index
            # we'll need this later when re-ordering
            indx_query_set[q_name] = one_query

            self.logger.debug(f"Running query: {q_name}",
                                 extra={
                                     'source': self.agent_name,
                                     'data': json.dumps(one_query, indent=4, cls=SafeEncoder)
                                 })

            if not enable:
                continue

            if not query:
                continue

            if query_mod:
                # we have to modify our query before passing to the LLM
                cannot_modify = False
                params = {}
                for col in query_mod.get("inputs", []):
                    if col not in extracted_info:
                        cannot_modify = True
                        break
                    else:
                        params[col] = extracted_info[col]
                if cannot_modify:
                    continue

                # modify the query using the params collected
                query = query.format(**params)

                # check if we need to modify alt queries also
                apply_to = query_mod.get("apply_to", "first")
                if apply_to == "all":
                    query_alts = [q.format(**params) for q in query_alts]

            # collect all the queries we need to run
            queries = [query] + query_alts

            # run the queries against the LLM
            for query in queries:

                # get the answer
                response = self.query(query, instructions=instructions, mode=query_mode)
                answer = response['answer']
                sources = response['sources']
                # normalized answer
                s_answer = answer.translate(str.maketrans('', '', string.punctuation)).lower().strip()

                # check the answer for UNKNOWN
                # but only in the case when postprocess->response==fill
                # this is for correct handling of spec extension/replacement
                if pp_response == 'fill':
                    if s_answer in retry_answers:
                        continue

                # post-process if needed
                # check if handler is callable
                if callable(pp_handler):
                    # collect all the columns needed to post-process the answer
                    params = {q_name: answer}
                    for col in pp_othercols:
                        if col in extracted_info:
                            if isinstance(extracted_info[col], list):
                                val = extracted_info[col][0]
                            else:
                                val = extracted_info[col]
                            params[col] = val
                    answer = pp_handler(params)

                # if next step action is 'extend'
                # then, we need to extend the query spec
                if pp_response == "extend":
                    query_set.extend(answer)
                    # continue, so that we move to the next query
                    continue

                # if next step action is 'replace'
                # then, we need to replace the query_set items in the default query spec
                # with the newly loaded one
                if pp_response == "replace":
                    # get the names of all query items in replace spec
                    replace_query_set_items = {}
                    for q_item in answer:
                        replace_query_set_items[q_item['name']] = q_item
                    # now, run through each query item in the default query set
                    # and check if replacement is needed
                    for i in range(len(query_set)):
                        if query_set[i]['name'] in replace_query_set_items:
                            # we have found a query item that needs to be replaced
                            query_set[i] = replace_query_set_items[query_set[i]['name']]

                    # continue, so that we move to the next query
                    continue

                # add the answers to the columns in the extracted dataset
                if len(fill_columns) == 0:
                    break
                if len(fill_columns) > 1:
                    i = 0
                    for col in fill_columns:
                        ans_i = answer[i]
                        ans = extracted_info.get(col,[])
                        if answer in ans:
                            # we have found this answer before
                            # no need to collect it again
                            continue
                        ans.append(ans_i)
                        extracted_info[col] = ans
                        i += 1
                else:
                    col = fill_columns[0] # have only one column to fill
                    ans = extracted_info.get(col,[])
                    if answer in ans:
                        # we have found this answer before
                        # no need to collect it again
                        continue
                    ans.append(answer)
                    extracted_info[col] = ans

                # at this point, at least one alt query has response
                # collect the grounding elements
                for col in fill_columns:
                    curr_sources = grounding.get(col, [])
                    curr_sources.append(sources)
                    grounding[col] = curr_sources

                # check if we need to run other alts
                if use_alts == "on-fail":
                    # no need to run an alt query
                    # since we have atleast some response
                    break

        # get the order of fields to return
        reordered_query_set = []
        field_order = spec.get("order")
        if not field_order:
            # we don't have an explicit order provided
            # use the default ordering
            reordered_query_set = query_set
        else:
            for field in field_order:
                if field in indx_query_set:
                    one_query = indx_query_set[field]
                    reordered_query_set.append(one_query)

        # check if all columns exist
        # and add the collected grounding
        extracts = {}
        for one_query in reordered_query_set:
            # check if we need to inlcude this field
            include = one_query.get("include", True)
            if not include:
                continue

            # we need to include this field
            fill_columns = one_query.get("fill_columns", [])
            for col in fill_columns:
                extracts[col] = {
                    "n_answers": 0 if not extracted_info.get(col) else len(extracted_info[col]),
                    "answer": extracted_info.get(col, []),
                    "sources": grounding.get(col, []),
                }

        return extracts

    def process_spec_signatures(self, spec):
        """
        check if signatures are present
        """
        self.logger.debug("Detecting signatures...",
                             extra={'source': self.agent_name})

        if spec.get("detect_signatures", False) == False:
            return None

        pages = []
        confidence = 0
        if len(self.doc_signatures) > 0:
            for signature in self.doc_signatures:
                pages.append(signature['page'])
                confidence += signature['confidence']
            n_sigs = len(pages) # this is correct
            pages = list(set(pages))
            n_pages = len(pages)
            confidence = round(confidence/n_sigs, 2)

            comment = f"Detected {n_sigs} signatures across {n_pages} pages"

            signatures = {
                "found": True,
                "comment": comment,
                "n_signatures": n_sigs,
                "n_pages": n_pages,
                "pages": pages,
                "confidence": confidence,
            }
        else:
            signatures = {
                "found": False,
                "comment": f"No signatures detected",
            }

        return signatures

    def process_spec(self, spec):
        """
        process a profilespec
        """
        name = spec.get("name")
        self.logger.debug(f"Processing spec: {name}",
                             extra={
                                 'source': self.agent_name,
                                 'data': json.dumps(spec, indent=4, cls=SafeEncoder)
                             })

        # check for signatures
        signatures = self.process_spec_signatures(spec)

        # process the questions
        extracts = self.process_spec_queries(spec)

        result = {
            "spec": name,
            "timestamp": time.time(),
            "extracts": extracts,
            "signatures": signatures
        }

        return result

    def extracts_to_df(self, info):
        """
        Convert extracts object into dataframe
        """
        extracts = info['extracts']
        entries = []

        for col, details in extracts.items():
            entry = {
                "field": col,
                "answer": details['answer'],
                "grounding": json.dumps(details['sources']),
            }
            entries.append(entry)

        df = pd.DataFrame(entries)

        return df


################### BEGIN TESTBED ###################

if __name__ == "__main__":

    # vars
    # cred = get_credentials_by_name('openai-api')
    persist_directory = "chromadb123"
    path = '...'

    agentname = "test_agent"
    # profilespec = get_profilespec()
    print(json.dumps(profilespec, indent=4, cls=SafeEncoder))
    platform = "azure"

    # create an agent
    agent = LLMQuerierExtractor(name=agentname, platform=platform)

    # load the data
    data = agent.load_data(content=path)

    # add to index
    agent.create_add_index(data=data,
                           store="chroma",
                           persist_directory=persist_directory,
                           index_name=agentname)

    # run the profilespec of queries
    info = agent.process_spec(profilespec)

    # output extraction
    for field, extract in info.get("extracts", {}).items():
        print (f"{field}: {extract.get('answer')}")



import os
import requests
from requests.auth import HTTPBasicAuth
import urllib.parse
import json
import importlib.util
import sys

from enrichsdk.utils import SafeEncoder

###############################
# Profile handling
###############################
def get_profile(clsobj, spec_category):

    profile_source = clsobj.args.get('profile_source', 'file')
    if profile_source == "api":
        is_valid, profile, l_msg = get_profile_from_api(clsobj, spec_category)
    elif profile_source == "file":
        is_valid, profile, l_msg = get_profile_from_file(clsobj)
    else:
        raise Exception("profile_source={api/file} must be specified")

    msg = f"Profile source: {profile_source}" + "\n"
    msg += l_msg
    msg += f"Profile: {json.dumps(profile, indent=4, cls=SafeEncoder)}" + "\n"

    return is_valid, profile, msg

###############################
# API based profile
###############################
def call_api(apicred, urlsuffix):
    msg = ""
    apikey = apicred.get('api_key', apicred.get('apikey'))
    htuser = apicred.get('basicauth', {}).get('user', "")
    htpass = apicred.get('basicauth', {}).get('pass', "")

    # Construct the url
    server = apicred.get('server', apicred.get('url'))
    if not server.startswith("http"):
        server = "https://" + server
    if server.endswith("/"):
        server = server[:-1]
    if urlsuffix.startswith("/"):
        urlsuffix = urlsuffix[1:]
    url = server + "/" + urlsuffix

    headers = {
        'accept': 'application/json',
        'X-API-Key': apikey,
    }

    msg += f"Calling URL: {url}" + "\n"
    if (htuser == '') and (htpass == ''):
        response = requests.get(url,
                                headers=headers,
                                verify=False,
                                timeout=10)
    else:
        response = requests.get(url,
                                headers=headers,
                                auth=HTTPBasicAuth(htuser, htpass),
                                verify=False,
                                timeout=10)
    is_valid = True if response.status_code == 200 else False
    msg += f"Response: {response.reason}" + "\n"

    return is_valid, response, msg


def load_profile_api(args, spec_category):
    msg = ""

    # API endpoint for anomalies service
    apicred = args['apicred']
    urlsuffix = f"/api/v2/dashboard/specs/"
    is_valid, response, l_msg = call_api(apicred, urlsuffix=urlsuffix)
    msg += l_msg
    if not is_valid:
        return is_valid, None, msg

    # now, loop through to get the app name
    app_name = None
    specs = []
    jdata = response.json()['data']
    for app_id, app_spec in jdata.items():
        if app_id == spec_category:
            for app in app_spec:
                app_name = app['name'] # for now, use the last app, generalize later

                # now, get the specs from the app
                app_name_url = spec_category.split('.')[-1]
                urlsuffix = f"/api/v2/app/{app_name_url}/{urllib.parse.quote(app_name)}/policies"
                is_valid, response, l_msg = call_api(apicred, urlsuffix)
                msg += l_msg
                if not is_valid:
                    continue
                    # return None, is_valid, msg

                specs += response.json()['data']

    profile = {
        "specs": specs
    }

    return is_valid, profile, msg

def get_profile_from_api(clsobj, spec_category):
    """
    Read the profile json from API
    """

    msg = ""

    if (not hasattr(clsobj, "args")):
        raise Exception(
            "'args' transform attribute should be defined"
        )
    for p in ['apicred']:
        if clsobj.args.get(p) == None:
            raise Exception(
                f"'{p}' attribute in args should be defined"
                )

    # call the API to get the anomaly specs
    msg += f"Loading profile from API" + "\n"
    is_valid, profile, l_msg = load_profile_api(clsobj.args, spec_category)
    msg += l_msg

    if 'specs' in profile:
        msg += f"Found {len(profile['specs'])} policies for spec: {spec_category}" + "\n"
    else:
        msg += f"Could not find 'specs' in the API response"

    return is_valid, profile, msg


###############################
# File based profile
###############################
def get_profile_from_file(clsobj):
    """
    Read the profile json from profilespec
    """

    is_valid = True
    msg = ""

    if (not hasattr(clsobj, "profiledir")) and (not hasattr(clsobj, "profilefile")):
        raise Exception(
            "'profiledir' transform attribute should be defined to use default get_profile method"
        )

    paths = []
    if hasattr(clsobj, "profilefile"):
        paths.append(self.profilefile)

    if hasattr(clsobj, "profiledir"):
        paths.extend(
            [
                clsobj.profiledir + "/profile.json",
                clsobj.profiledir + "/profile.yaml",
                clsobj.profiledir + "/profilespec.json",
                clsobj.profiledir + "/profilespec.yaml",
                clsobj.profiledir + "/profilespec.py",
            ]
        )

    profile = None
    for p in paths:
        if not os.path.exists(p):
            continue
        if p.endswith(".json"):
            profile = json.load(open(p))
        elif p.endswith(".yaml"):
            profile = yaml.load(open(p))
        elif p.endswith(".py"):
            # => Load from the file...
            libname = str(clsobj) + "_profile"
            spec = importlib.util.spec_from_file_location(libname, p)
            foo = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(foo)
            if hasattr(foo, 'get_profile'):
                profile = foo.get_profile()
            elif hasattr(foo, 'get_profile_spec'):
                profile = foo.get_profile_spec()
            elif hasattr(foo, 'profile'):
                profile = foo.profile

    if profile is None:
        raise Exception("Profile could not be found")

    return is_valid, profile, msg


###############################
# DB handlers
###############################


###############################
# Dataset handlers
###############################
def construct_dataset_list(clsobj, specs):
    if not hasattr(clsobj, 'get_dataset_registry'):
        raise Exception(
            "get_datasets expects get_dataset_registry method"
        )

    # call the overloaded method to get the dataset registry
    registry = clsobj.get_dataset_registry()

    # what are all the datasets in the spec
    spec_datasets = []
    for spec in specs:
        type = spec.get('config', {}).get('source', {}).get('type')
        dataset = spec.get('config', {}).get('source', {}).get('dataset')
        if type != 'registry':
            continue
        if dataset == None:
            continue
        spec_datasets.append(dataset)

    # iterate to keep only the datasets in the spec
    datasets = {}
    for dataset in registry.datasets:
        name = dataset.name
        for subset in dataset.subsets:
            d = f"{name}-{subset['name']}"
            if d in spec_datasets:
                # make a lookup table
                datasets[d] = dataset

    return datasets



import os
import sys
import json
import copy
import logging
import traceback
import humanize
from datetime import datetime
from dateutil import parser as date_parser
import logging

logger = logging.getLogger("app")

###############################################################
# Run History
###############################################################
class Run(object):
    """
    A single instance of command run. This class will mediate
    access to the command logs and metadata.

    """

    def __init__(self, rundir, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rundir = rundir
        self.runid = os.path.basename(self.rundir)
        epoch = datetime.fromtimestamp(0)
        self.start_time = epoch
        self.end_time = epoch

        try:
            filename = os.path.join(self.rundir, "metadata.json")
            self.metadata = json.load(open(filename))
            schema = self.metadata.get("schema", "unknown")
            command = self.metadata.get("command", {})

            for attr in [
                "host",
                "name",
                "status",
                "description",
                "cmdline",
                "message",
                "user",
                "runid",
                "start_time",
                "end_time",
            ]:

                if schema == "standalone:command":
                    if attr in command:
                        if attr in ["start_time", "end_time"]:
                            value = date_parser.parse(command[attr])
                        else:
                            value = command[attr]
                        setattr(self, attr, value)
                    elif attr == "user":
                        value = command["stats"]["user"]
                        setattr(self, attr, value)
                    elif attr == "user":
                        value = command["stats"]["user"]
                        setattr(self, attr, value)
                else:
                    if attr in ["start_time", "end_time"]:
                        value = date_parser.parse(self.metadata[attr])
                    elif attr == "user":
                        value = self.metadata.get("username", "unknown")
                    elif attr == "name":
                        value = self.metadata.get("command", "unknown")
                    else:
                        value = self.metadata.get(attr, "unknown")

                    # Rename one attribute
                    setattr(self, attr, value)

            self.outputs = self.metadata.get("outputs", [])
        except:
            traceback.print_exc()
            for attr in [
                "host",
                "name",
                "status",
                "description",
                "cmdline" "message",
                "user",
            ]:
                setattr(self, attr, "unknown")
            self.outputs = []

        logfile = self.metadata.get("log", "log.json")
        self.logfile = os.path.join(self.rundir, logfile)
        self.log = open(self.logfile).readlines()

        self.duration = (self.end_time - self.start_time).total_seconds()

    def __str__(self):
        return self.name

    @property
    def username(self):
        return self.user

    def get_command(self):
        return self.name.replace("enrichcmd run ", "").strip()

    def get_cmdline(self):
        cmdline = copy.copy(self.cmdline)
        if cmdline[0].endswith("enrichcmd"):
            cmdline[0] = "enrichcmd"
        return " ".join(cmdline)


class RunManager(object):
    """
    Command Run Management to enable access to
    pass runs of the commands.

    """

    def __init__(self, conf, *args, **kwargs):
        """

        Args:
            conf (dict): Only one parameter supported now 'command_root'
        """
        super().__init__(*args, **kwargs)
        self.runs = []
        self.conf = conf

        if "command_root" in conf:
            command_roots = [conf["command_root"]]
        elif "command_roots" in conf:
            command_roots = conf["command_roots"]
        else:
            raise Exception(
                "Configuration should specify command_root or command_roots parameters"
            )

        self.command_roots = []
        for p in command_roots:
            self.command_roots.append(os.path.abspath(p))

        # Collect rundirs from all directories...
        rundirs = []
        for p in self.command_roots:
            if not os.path.exists(p):
                logger.debug("Invalid command root directory: {}".format(p))
                continue

            rundirs.extend(
                [{"name": d, "path": os.path.join(p, d)} for d in os.listdir(p)]
            )

        rundirs = sorted(rundirs, key=lambda r: r["name"], reverse=True)

        count = conf.get("history", 100)
        rundirs = rundirs[:count]
        errors = []
        for r in rundirs:
            try:
                runid = r["name"]
                rundir = r["path"]
                self.runs.append(Run(rundir))
            except:
                traceback.print_exc()
                errors.append(runid)

        if len(errors) > 0:
            logger.debug(
                "Found errors while loading command history",
                extra={"transform": "Commands", "data": str(errors)},
            )

    def get_runs(self, username=None):
        """
        Get runs for a given username
        """

        runs = [r for r in self.runs if ((username is None) or (r.user == username))]
        return runs

    def get_run(self, runid):

        run = None
        for p in self.command_roots:
            rundir = os.path.join(p, runid)
            if os.path.exists(rundir):
                try:
                    run = Run(rundir)
                except:
                    pass

        return run

    def show_history(self, n=10):

        runs = sorted(self.runs, key=lambda r: r.start_time, reverse=True)

        template = "{:10s}{:30s}{:10s}{:15s}{:20s}"
        print(template.format("Who", "Description", "Status", "When", "Name"))
        for r in runs[:n]:
            print(
                template.format(
                    r.user,
                    r.description[:27],
                    r.status[:10],
                    humanize.naturaltime(r.start_time),
                    r.name,
                )
            )



from enrichsdk.app.utils import generate_id
import json
import logging


class BaseWidget:
    def __init__(self, widgetid, name, title, template):

        if name is None:
            raise Exception("{} can't be null for the widget".format("Name"))

        if template is None:
            raise Exception("{} can't be null for the widget".format("Template"))

        self.name = name
        self.title = title
        self.type = template

        if widgetid is None:
            self.id = generate_id(self.name)
        else:
            self.id = widgetid


    def to_json(self):
        return json.dumps(self.__dict__)






from django.conf.urls import url, include
from . import views, catalog, custom
from dashboard.app import *

app_name = "APPNAME"

# Notes:
# 1. This merges the urls with the rest of the application URLs. Please be careful. 
# 2. app_urlpatterns is only for enrichapp.dashboard.custom app. All other apps ignore this field.
# 3. The reusable app and the app integrating the reusable app are different. Having both in the same urls.py will result in unexpected behavior

urlpatterns = []

# => Base that will be imported
APPNAME_urlpatterns = []

# Custom app
customspec = custom.get_spec()
register_app_instance(
    appmod="enrichapp.dashboard.custom",
    namespace="default",
    rootnamespace=app_name,
    urlpatterns=urlpatterns,
    spec=customspec,
    app_urlpatterns=[
        url(r'^[/]?$', views.index, name="index"),
    ]
)

catalogspec = catalog.get_spec()
register_app_instance(
    appmod="enrichapp.dashboard.catalog",
    namespace="catalog",
    rootnamespace=app_name,
    urlpatterns=urlpatterns,
    spec=catalogspec)

#from . import persona
#searchspec = persona.get_spec()
#register_app_instance(
#    appmod="enrichapp.dashboard.persona",
#    namespace="persona",
#    rootnamespace=app_name,
#    urlpatterns=urlpatterns,
#    spec=searchspec)
#
